#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import time
import pathlib
import logging
import json
import copy
import pickle
from datetime import datetime, timezone
from typing import Optional, Tuple
from importlib.metadata import version
import numpy as np
import pandas as pd

from distutils.util import strtobool

from emhass.retrieve_hass import RetrieveHass
from emhass.forecast import Forecast
from emhass.machine_learning_forecaster import MLForecaster
from emhass.optimization import Optimization
from emhass.machine_learning_regressor import MLRegressor
from emhass import utils


def set_input_data_dict(emhass_conf: dict, costfun: str,
                        params: str, runtimeparams: str, set_type: str, logger: logging.Logger,
                        get_data_from_file: Optional[bool] = False) -> dict:
    """
    Set up some of the data needed for the different actions.

    :param emhass_conf: Dictionary containing the needed emhass paths
    :type emhass_conf: dict
    :param costfun: The type of cost function to use for optimization problem
    :type costfun: str
    :param params: Configuration parameters passed from data/options.json
    :type params: str
    :param runtimeparams: Runtime optimization parameters passed as a dictionary
    :type runtimeparams: str
    :param set_type: Set the type of setup based on following type of optimization
    :type set_type: str
    :param logger: The passed logger object
    :type logger: logging object
    :param get_data_from_file: Use data from saved CSV file (useful for debug)
    :type get_data_from_file: bool, optional
    :return: A dictionnary with multiple data used by the action functions
    :rtype: dict

    """
    logger.info("Setting up needed data")
    # Parsing yaml
    retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(
        emhass_conf, use_secrets=not(get_data_from_file), params=params)
    # Treat runtimeparams
    params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
        runtimeparams, params, retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
    # Define main objects
    rh = RetrieveHass(retrieve_hass_conf['hass_url'], retrieve_hass_conf['long_lived_token'],
                      retrieve_hass_conf['freq'], retrieve_hass_conf['time_zone'],
                      params, emhass_conf, logger, get_data_from_file=get_data_from_file)
    fcst = Forecast(retrieve_hass_conf, optim_conf, plant_conf,
                    params, emhass_conf, logger, get_data_from_file=get_data_from_file)
    opt = Optimization(retrieve_hass_conf, optim_conf, plant_conf,
                       fcst.var_load_cost, fcst.var_prod_price,
                       costfun, emhass_conf, logger)
    # Perform setup based on type of action
    if set_type == "perfect-optim":
        # Retrieve data from hass
        if get_data_from_file:
            with open(emhass_conf['data_path'] / 'test_df_final.pkl', 'rb') as inp:
                rh.df_final, days_list, var_list = pickle.load(inp)
            retrieve_hass_conf['var_load'] = str(var_list[0])
            retrieve_hass_conf['var_PV'] = str(var_list[1])
            retrieve_hass_conf['var_interp'] = [
                retrieve_hass_conf['var_PV'], retrieve_hass_conf['var_load']]
            retrieve_hass_conf['var_replace_zero'] = [
                retrieve_hass_conf['var_PV']]
        else:
            days_list = utils.get_days_list(
                retrieve_hass_conf["days_to_retrieve"])
            var_list = [retrieve_hass_conf["var_load"],
                        retrieve_hass_conf["var_PV"]]
            if not rh.get_data(days_list, var_list, minimal_response=False, significant_changes_only=False):
                return False
        if not rh.prepare_data(retrieve_hass_conf["var_load"],
                               load_negative=retrieve_hass_conf["load_negative"],
                               set_zero_min=retrieve_hass_conf["set_zero_min"],
                               var_replace_zero=retrieve_hass_conf["var_replace_zero"],
                               var_interp=retrieve_hass_conf["var_interp"]):
            return False
        df_input_data = rh.df_final.copy()
        # What we don't need for this type of action
        P_PV_forecast, P_load_forecast, df_input_data_dayahead = None, None, None
    elif set_type == "dayahead-optim":
        # Get PV and load forecasts
        df_weather = fcst.get_weather_forecast(
            method=optim_conf["weather_forecast_method"])
        P_PV_forecast = fcst.get_power_from_weather(df_weather)
        P_load_forecast = fcst.get_load_forecast(
            method=optim_conf['load_forecast_method'])
        if isinstance(P_load_forecast, bool) and not P_load_forecast:
            logger.error(
                "Unable to get sensor power photovoltaics, or sensor power load no var loads. Check HA sensors and their daily data")
            return False
        df_input_data_dayahead = pd.DataFrame(np.transpose(np.vstack(
            [P_PV_forecast.values, P_load_forecast.values])), index=P_PV_forecast.index,
            columns=["P_PV_forecast", "P_load_forecast"])
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)
        params = json.loads(params)
        if ("prediction_horizon" in params["passed_data"] and params["passed_data"]["prediction_horizon"] is not None):
            prediction_horizon = params["passed_data"]["prediction_horizon"]
            df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[
                df_input_data_dayahead.index[0]: df_input_data_dayahead.index[prediction_horizon - 1]]
        # What we don't need for this type of action
        df_input_data, days_list = None, None
    elif set_type == "naive-mpc-optim":
        # Retrieve data from hass
        if get_data_from_file:
            with open(emhass_conf['data_path'] / 'test_df_final.pkl', 'rb') as inp:
                rh.df_final, days_list, var_list = pickle.load(inp)
            retrieve_hass_conf['var_load'] = str(var_list[0])
            retrieve_hass_conf['var_PV'] = str(var_list[1])
            retrieve_hass_conf['var_interp'] = [
                retrieve_hass_conf['var_PV'], retrieve_hass_conf['var_load']]
            retrieve_hass_conf['var_replace_zero'] = [
                retrieve_hass_conf['var_PV']]
        else:
            days_list = utils.get_days_list(1)
            var_list = [retrieve_hass_conf["var_load"],
                        retrieve_hass_conf["var_PV"]]
            if not rh.get_data(days_list, var_list, minimal_response=False, significant_changes_only=False):
                return False
        if not rh.prepare_data(retrieve_hass_conf["var_load"],
                               load_negative=retrieve_hass_conf["load_negative"],
                               set_zero_min=retrieve_hass_conf["set_zero_min"],
                               var_replace_zero=retrieve_hass_conf["var_replace_zero"],
                               var_interp=retrieve_hass_conf["var_interp"]):
            return False
        df_input_data = rh.df_final.copy()
        # Get PV and load forecasts
        df_weather = fcst.get_weather_forecast(
            method=optim_conf['weather_forecast_method'])
        P_PV_forecast = fcst.get_power_from_weather(
            df_weather, set_mix_forecast=True, df_now=df_input_data)
        P_load_forecast = fcst.get_load_forecast(
            method=optim_conf['load_forecast_method'], set_mix_forecast=True, df_now=df_input_data)
        if isinstance(P_load_forecast, bool) and not P_load_forecast:
            logger.error(
                "Unable to get sensor power photovoltaics, or sensor power load no var loads. Check HA sensors and their daily data")
            return False
        df_input_data_dayahead = pd.concat([P_PV_forecast, P_load_forecast], axis=1)
        df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)
        df_input_data_dayahead.columns = ["P_PV_forecast", "P_load_forecast"]
        params = json.loads(params)
        if ("prediction_horizon" in params["passed_data"] and params["passed_data"]["prediction_horizon"] is not None):
            prediction_horizon = params["passed_data"]["prediction_horizon"]
            df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[
                df_input_data_dayahead.index[0]: df_input_data_dayahead.index[prediction_horizon - 1]]
    elif (set_type == "forecast-model-fit" or set_type == "forecast-model-predict" or set_type == "forecast-model-tune"):
        df_input_data_dayahead = None
        P_PV_forecast, P_load_forecast = None, None
        params = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = params["passed_data"]["days_to_retrieve"]
        model_type = params["passed_data"]["model_type"]
        var_model = params["passed_data"]["var_model"]
        if get_data_from_file:
            days_list = None
            filename = 'data_train_'+model_type+'.pkl'
            filename_path = emhass_conf['data_path'] / filename
            with open(filename_path, 'rb') as inp:
                df_input_data, _ = pickle.load(inp)
            df_input_data = df_input_data[df_input_data.index[-1] - pd.offsets.Day(days_to_retrieve):]
        else:
            days_list = utils.get_days_list(days_to_retrieve)
            var_list = [var_model]
            if not rh.get_data(days_list, var_list):
                return False
            df_input_data = rh.df_final.copy()
    elif set_type == "regressor-model-fit" or set_type == "regressor-model-predict":
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        params = json.loads(params)
        days_list = None
        csv_file = params["passed_data"].get("csv_file", None)
        if "features" in params["passed_data"]:
            features = params["passed_data"]["features"]
        if "target" in params["passed_data"]:
            target = params["passed_data"]["target"]
        if "timestamp" in params["passed_data"]:
            timestamp = params["passed_data"]["timestamp"]
        if csv_file:
            if get_data_from_file:
                base_path = emhass_conf["data_path"]  # + "/data"
                filename_path = pathlib.Path(base_path) / csv_file
            else:
                filename_path = emhass_conf["data_path"] / csv_file
            if filename_path.is_file():
                df_input_data = pd.read_csv(filename_path, parse_dates=True)
            else:
                logger.error("The CSV file " + csv_file +
                             " was not found in path: " + str(emhass_conf["data_path"]))
                return False
                # raise ValueError("The CSV file " + csv_file + " was not found.")
            required_columns = []
            required_columns.extend(features)
            required_columns.append(target)
            if timestamp is not None:
                required_columns.append(timestamp)
            if not set(required_columns).issubset(df_input_data.columns):
                logger.error(
                    "The cvs file does not contain the required columns.")
                msg = f"CSV file should contain the following columns: {', '.join(required_columns)}"
                logger.error(msg)
                return False
    elif set_type == "publish-data":
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        days_list = None
    else:
        logger.error(
            "The passed action argument and hence the set_type parameter for setup is not valid",
        )
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        days_list = None
    # The input data dictionary to return
    input_data_dict = {
        'emhass_conf': emhass_conf,
        'retrieve_hass_conf': retrieve_hass_conf,
        'rh': rh,
        'opt': opt,
        'fcst': fcst,
        'df_input_data': df_input_data,
        'df_input_data_dayahead': df_input_data_dayahead,
        'P_PV_forecast': P_PV_forecast,
        'P_load_forecast': P_load_forecast,
        'costfun': costfun,
        'params': params,
        'days_list': days_list
    }
    return input_data_dict


def perfect_forecast_optim(input_data_dict: dict, logger: logging.Logger,
                           save_data_to_file: Optional[bool] = True, 
                           debug: Optional[bool] = False) -> pd.DataFrame:
    """
    Perform a call to the perfect forecast optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing perfect forecast optimization")
    # Load cost and prod price forecast
    df_input_data = input_data_dict['fcst'].get_load_cost_forecast(
        input_data_dict['df_input_data'],
        method=input_data_dict['fcst'].optim_conf['load_cost_forecast_method'],
        list_and_perfect=True)
    if isinstance(df_input_data, bool) and not df_input_data:
        return False
    df_input_data = input_data_dict['fcst'].get_prod_price_forecast(
        df_input_data, method=input_data_dict['fcst'].optim_conf['prod_price_forecast_method'],
        list_and_perfect=True)
    if isinstance(df_input_data, bool) and not df_input_data:
        return False
    opt_res = input_data_dict['opt'].perform_perfect_forecast_optim(
        df_input_data, input_data_dict['days_list'])
    # Save CSV file for analysis
    if save_data_to_file:
        filename = "opt_res_perfect_optim_" + \
            input_data_dict["costfun"] + ".csv"
    else:  # Just save the latest optimization results
        filename = "opt_res_latest.csv"
    if not debug:
        opt_res.to_csv(
            input_data_dict['emhass_conf']['data_path'] / filename, index_label='timestamp')
        

    if not isinstance(input_data_dict["params"],dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    # if continual_publish, save perfect results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish",False) or params["passed_data"].get("entity_save",False):
        #Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)   

    return opt_res


def dayahead_forecast_optim(input_data_dict: dict, logger: logging.Logger, 
                            save_data_to_file: Optional[bool] = False, 
                            debug: Optional[bool] = False) -> pd.DataFrame:
    """
    Perform a call to the day-ahead optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing day-ahead forecast optimization")
    # Load cost and prod price forecast
    df_input_data_dayahead = input_data_dict['fcst'].get_load_cost_forecast(
        input_data_dict['df_input_data_dayahead'],
        method=input_data_dict['fcst'].optim_conf['load_cost_forecast_method'])
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    df_input_data_dayahead = input_data_dict['fcst'].get_prod_price_forecast(
        df_input_data_dayahead,
        method=input_data_dict['fcst'].optim_conf['prod_price_forecast_method'])
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    opt_res_dayahead = input_data_dict['opt'].perform_dayahead_forecast_optim(
        df_input_data_dayahead, input_data_dict['P_PV_forecast'], input_data_dict['P_load_forecast'])
    # Save CSV file for publish_data
    if save_data_to_file:
        today = datetime.now(timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
    else:  # Just save the latest optimization results
        filename = "opt_res_latest.csv"
    if not debug:
        opt_res_dayahead.to_csv(
          input_data_dict['emhass_conf']['data_path'] / filename, index_label='timestamp')
    
    if not isinstance(input_data_dict["params"],dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    
    # if continual_publish, save day_ahead results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish",False) or params["passed_data"].get("entity_save",False):
        #Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)   
        
    return opt_res_dayahead


def naive_mpc_optim(input_data_dict: dict, logger: logging.Logger, 
                    save_data_to_file: Optional[bool] = False, 
                    debug: Optional[bool] = False) -> pd.DataFrame:
    """
    Perform a call to the naive Model Predictive Controller optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing naive MPC optimization")
    # Load cost and prod price forecast
    df_input_data_dayahead = input_data_dict['fcst'].get_load_cost_forecast(
        input_data_dict['df_input_data_dayahead'],
        method=input_data_dict['fcst'].optim_conf['load_cost_forecast_method'])
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    df_input_data_dayahead = input_data_dict['fcst'].get_prod_price_forecast(
        df_input_data_dayahead, method=input_data_dict['fcst'].optim_conf['prod_price_forecast_method'])
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    # The specifics params for the MPC at runtime
    prediction_horizon = input_data_dict["params"]["passed_data"]["prediction_horizon"]
    soc_init = input_data_dict["params"]["passed_data"]["soc_init"]
    soc_final = input_data_dict["params"]["passed_data"]["soc_final"]
    def_total_hours = input_data_dict["params"]["passed_data"]["def_total_hours"]
    def_start_timestep = input_data_dict["params"]["passed_data"]["def_start_timestep"]
    def_end_timestep = input_data_dict["params"]["passed_data"]["def_end_timestep"]
    opt_res_naive_mpc = input_data_dict["opt"].perform_naive_mpc_optim(
        df_input_data_dayahead, input_data_dict["P_PV_forecast"], input_data_dict["P_load_forecast"],
        prediction_horizon, soc_init, soc_final, def_total_hours,
        def_start_timestep, def_end_timestep)
    # Save CSV file for publish_data
    if save_data_to_file:
        today = datetime.now(timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        filename = "opt_res_naive_mpc_" + today.strftime("%Y_%m_%d") + ".csv"
    else:  # Just save the latest optimization results
        filename = "opt_res_latest.csv"
    if not debug:
        opt_res_naive_mpc.to_csv(
          input_data_dict['emhass_conf']['data_path'] / filename, index_label='timestamp')
        
    if not isinstance(input_data_dict["params"],dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    # if continual_publish, save mpc results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish",False) or params["passed_data"].get("entity_save",False):
        #Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)   

    return opt_res_naive_mpc


def forecast_model_fit(input_data_dict: dict, logger: logging.Logger, 
                       debug: Optional[bool] = False) -> Tuple[pd.DataFrame, pd.DataFrame, MLForecaster]:
    """Perform a forecast model fit from training data retrieved from Home Assistant.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :return: The DataFrame containing the forecast data results without and with backtest and the `mlforecaster` object
    :rtype: Tuple[pd.DataFrame, pd.DataFrame, mlforecaster]
    """
    data = copy.deepcopy(input_data_dict['df_input_data'])
    model_type = input_data_dict['params']['passed_data']['model_type']
    var_model = input_data_dict['params']['passed_data']['var_model']
    sklearn_model = input_data_dict['params']['passed_data']['sklearn_model']
    num_lags = input_data_dict['params']['passed_data']['num_lags']
    split_date_delta = input_data_dict['params']['passed_data']['split_date_delta']
    perform_backtest = input_data_dict['params']['passed_data']['perform_backtest']
    # The ML forecaster object
    mlf = MLForecaster(data, model_type, var_model, sklearn_model,
                       num_lags, input_data_dict['emhass_conf'], logger)
    # Fit the ML model
    df_pred, df_pred_backtest = mlf.fit(
        split_date_delta=split_date_delta, perform_backtest=perform_backtest
    )
    # Save model
    if not debug:
        filename = model_type+'_mlf.pkl'
        filename_path = input_data_dict['emhass_conf']['data_path'] / filename
        with open(filename_path, 'wb') as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
    return df_pred, df_pred_backtest, mlf


def forecast_model_predict(input_data_dict: dict, logger: logging.Logger, 
                           use_last_window: Optional[bool] = True, 
                           debug: Optional[bool] = False, mlf: Optional[MLForecaster] = None
                           ) -> pd.DataFrame:
    r"""Perform a forecast model predict using a previously trained skforecast model.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param use_last_window: True if the 'last_window' option should be used for the \
        custom machine learning forecast model. The 'last_window=True' means that the data \
        that will be used to generate the new forecast will be freshly retrieved from \
        Home Assistant. This data is needed because the forecast model is an auto-regressive \
        model with lags. If 'False' then the data using during the model train is used. Defaults to True
    :type use_last_window: Optional[bool], optional
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
        and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
    :type mlf: Optional[mlforecaster], optional
    :return: The DataFrame containing the forecast prediction data
    :rtype: pd.DataFrame
    """
    # Load model
    model_type = input_data_dict['params']['passed_data']['model_type']
    filename = model_type+'_mlf.pkl'
    filename_path = input_data_dict['emhass_conf']['data_path'] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
        else:
            logger.error(
                "The ML forecaster file was not found, please run a model fit method before this predict method",
            )
            return
    # Make predictions
    if use_last_window:
        data_last_window = copy.deepcopy(input_data_dict["df_input_data"])
    else:
        data_last_window = None
    predictions = mlf.predict(data_last_window)
    # Publish data to a Home Assistant sensor
    model_predict_publish = input_data_dict["params"]["passed_data"][
        "model_predict_publish"
    ]
    model_predict_entity_id = input_data_dict["params"]["passed_data"][
        "model_predict_entity_id"
    ]
    model_predict_unit_of_measurement = input_data_dict["params"]["passed_data"][
        "model_predict_unit_of_measurement"
    ]
    model_predict_friendly_name = input_data_dict["params"]["passed_data"][
        "model_predict_friendly_name"
    ]
    publish_prefix = input_data_dict["params"]["passed_data"]["publish_prefix"]
    if model_predict_publish is True:
        # Estimate the current index
        now_precise = datetime.now(
            input_data_dict["retrieve_hass_conf"]["time_zone"]
        ).replace(second=0, microsecond=0)
        if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
            idx_closest = predictions.index.get_indexer([now_precise], method="nearest")[0]
        elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
            idx_closest = predictions.index.get_indexer([now_precise], method="ffill")[0]
        elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
            idx_closest = predictions.index.get_indexer([now_precise], method="bfill")[0]
        if idx_closest == -1:
            idx_closest = predictions.index.get_indexer([now_precise], method="nearest")[0]
        # Publish Load forecast
        input_data_dict["rh"].post_data(
            predictions, idx_closest, model_predict_entity_id,
            model_predict_unit_of_measurement, model_predict_friendly_name,
            type_var="mlforecaster", publish_prefix=publish_prefix)
    return predictions


def forecast_model_tune(input_data_dict: dict, logger: logging.Logger, 
                        debug: Optional[bool] = False, mlf: Optional[MLForecaster] = None
                        ) -> Tuple[pd.DataFrame, MLForecaster]:
    """Tune a forecast model hyperparameters using bayesian optimization.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
        and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
    :type mlf: Optional[mlforecaster], optional
    :return: The DataFrame containing the forecast data results using the optimized model
    :rtype: pd.DataFrame
    """
    # Load model
    model_type = input_data_dict['params']['passed_data']['model_type']
    filename = model_type+'_mlf.pkl'
    filename_path = input_data_dict['emhass_conf']['data_path'] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
        else:
            logger.error(
                "The ML forecaster file was not found, please run a model fit method before this tune method",
            )
            return None, None
    # Tune the model
    df_pred_optim = mlf.tune(debug=debug)
    # Save model
    if not debug:
        filename = model_type+'_mlf.pkl'
        filename_path = input_data_dict['emhass_conf']['data_path'] / filename
        with open(filename_path, 'wb') as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
    return df_pred_optim, mlf


def regressor_model_fit(input_data_dict: dict, logger: logging.Logger, 
                        debug: Optional[bool] = False) -> MLRegressor:
    """Perform a forecast model fit from training data retrieved from Home Assistant.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    """
    data = copy.deepcopy(input_data_dict["df_input_data"])
    if "model_type" in input_data_dict["params"]["passed_data"]:
        model_type = input_data_dict["params"]["passed_data"]["model_type"]
    else:
        logger.error("parameter: 'model_type' not passed")
        return False
    if "regression_model" in input_data_dict["params"]["passed_data"]:
        regression_model = input_data_dict["params"]["passed_data"]["regression_model"]
    else:
        logger.error("parameter: 'regression_model' not passed")
        return False
    if "features" in input_data_dict["params"]["passed_data"]:
        features = input_data_dict["params"]["passed_data"]["features"]
    else:
        logger.error("parameter: 'features' not passed")
        return False
    if "target" in input_data_dict["params"]["passed_data"]:
        target = input_data_dict["params"]["passed_data"]["target"]
    else:
        logger.error("parameter: 'target' not passed")
        return False
    if "timestamp" in input_data_dict["params"]["passed_data"]:
        timestamp = input_data_dict["params"]["passed_data"]["timestamp"]
    else:
        logger.error("parameter: 'timestamp' not passed")
        return False
    if "date_features" in input_data_dict["params"]["passed_data"]:
        date_features = input_data_dict["params"]["passed_data"]["date_features"]
    else:
        logger.error("parameter: 'date_features' not passed")
        return False
    # The MLRegressor object
    mlr = MLRegressor(data, model_type, regression_model, features, target, timestamp, logger)
    # Fit the ML model
    mlr.fit(date_features=date_features)
    # Save model
    if not debug:
        filename = model_type + "_mlr.pkl"
        filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
        with open(filename_path, "wb") as outp:
            pickle.dump(mlr, outp, pickle.HIGHEST_PROTOCOL)
    return mlr


def regressor_model_predict(input_data_dict: dict, logger: logging.Logger, 
                            debug: Optional[bool] = False, mlr: Optional[MLRegressor] = None
                            ) -> np.ndarray:
    """Perform a prediction from csv file.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    """
    if "model_type" in input_data_dict["params"]["passed_data"]:
        model_type = input_data_dict["params"]["passed_data"]["model_type"]
    else:
        logger.error("parameter: 'model_type' not passed")
        return False
    filename = model_type + "_mlr.pkl"
    filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlr = pickle.load(inp)
        else:
            logger.error(
                "The ML forecaster file was not found, please run a model fit method before this predict method",
            )
            return False
    if "new_values" in input_data_dict["params"]["passed_data"]:
        new_values = input_data_dict["params"]["passed_data"]["new_values"]
    else:
        logger.error("parameter: 'new_values' not passed")
        return False
    # Predict from csv file
    prediction = mlr.predict(new_values)
    mlr_predict_entity_id = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_entity_id", "sensor.mlr_predict")
    mlr_predict_unit_of_measurement = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_unit_of_measurement", "h")
    mlr_predict_friendly_name = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_friendly_name", "mlr predictor")
    # Publish prediction
    idx = 0
    if not debug:
        input_data_dict["rh"].post_data(prediction, idx, mlr_predict_entity_id,
            mlr_predict_unit_of_measurement, mlr_predict_friendly_name,
            type_var="mlregressor")
    return prediction

def publish_data(input_data_dict: dict, logger: logging.Logger, 
                 save_data_to_file: Optional[bool] = False, 
                 opt_res_latest: Optional[pd.DataFrame] = None, 
                 entity_save: Optional[bool] = False,
                 dont_post: Optional[bool] = False) -> pd.DataFrame:
    """
    Publish the data obtained from the optimization results.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: If True we will read data from optimization results in dayahead CSV file
    :type save_data_to_file: bool, optional
    :return: The output data of the optimization readed from a CSV file in the data folder
    :rtype: pd.DataFrame
    :param entity_save: Save built entities to data_path/entities
    :type entity_save: bool, optional
    :param dont_post: Do not post to Home Assistant. Works with entity_save
    :type dont_post: bool, optional

    """
    logger.info("Publishing data to HASS instance")
    if not isinstance(input_data_dict["params"],dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]
    # Check if a day ahead optimization has been performed (read CSV file)
    if save_data_to_file:
        today = datetime.now(timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
    # If publish_prefix is passed, check if there is saved entities in data_path/entities with prefix, publish to results
    elif params["passed_data"].get("publish_prefix","") != "" and not dont_post:
        opt_res_list = []
        opt_res_list_names = []
        publish_prefix = params["passed_data"]["publish_prefix"]
        entity_path = input_data_dict['emhass_conf']['data_path'] / "entities"
        # Check if items in entity_path
        if os.path.exists(entity_path) and len(os.listdir(entity_path)) > 0:
            # Obtain all files in entity_path
            entity_path_contents =  os.listdir(entity_path)    
            for entity in entity_path_contents:
                if entity != "metadata.json": 
                    # If publish_prefix is "all" publish all saved entities to Home Assistant 
                    # If publish_prefix matches the prefix from saved entities, publish to Home Assistant
                    if publish_prefix in entity or publish_prefix == "all":
                        entity_data = publish_json(entity,input_data_dict,entity_path,logger)    
                        if not isinstance(entity_data, bool):
                            opt_res_list.append(entity_data)
                            opt_res_list_names.append(entity.replace(".json", ""))
                        else:
                            return False      
            # Build a DataFrame with published entities
            opt_res = pd.concat(opt_res_list, axis=1)
            opt_res.columns = opt_res_list_names
            return opt_res
        else:
            logger.warning("no saved entity json files in path:" + str(entity_path))     
            logger.warning("falling back to opt_res_latest")
            filename = "opt_res_latest.csv"            
    else:
        filename = "opt_res_latest.csv"
    if opt_res_latest is None:
        if not os.path.isfile(input_data_dict['emhass_conf']['data_path'] / filename):
            logger.error(
                "File not found error, run an optimization task first.")
            return
        else:
            opt_res_latest = pd.read_csv(
                input_data_dict['emhass_conf']['data_path'] / filename, index_col='timestamp')
            opt_res_latest.index = pd.to_datetime(opt_res_latest.index)
            opt_res_latest.index.freq = input_data_dict["retrieve_hass_conf"]["freq"]
    # Estimate the current index
    now_precise = datetime.now(
        input_data_dict["retrieve_hass_conf"]["time_zone"]
    ).replace(second=0, microsecond=0)
    if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="nearest")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
        idx_closest = opt_res_latest.index.get_indexer(
            [now_precise], method="ffill")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
        idx_closest = opt_res_latest.index.get_indexer(
            [now_precise], method="bfill")[0]
    if idx_closest == -1:
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="nearest")[0]
    # Publish the data
    publish_prefix = params["passed_data"]["publish_prefix"]
    # Publish PV forecast
    custom_pv_forecast_id = params["passed_data"]["custom_pv_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_PV"],
        idx_closest,
        custom_pv_forecast_id["entity_id"],
        custom_pv_forecast_id["unit_of_measurement"],
        custom_pv_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    # Publish Load forecast
    custom_load_forecast_id = params["passed_data"]["custom_load_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_Load"],
        idx_closest,
        custom_load_forecast_id["entity_id"],
        custom_load_forecast_id["unit_of_measurement"],
        custom_load_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = ["P_PV", "P_Load"]
    # Publish PV curtailment
    custom_pv_curtailment_id = params["passed_data"]["custom_pv_curtailment_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_PV_curtailment"],
        idx_closest,
        custom_pv_curtailment_id["entity_id"],
        custom_pv_curtailment_id["unit_of_measurement"],
        custom_pv_curtailment_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = cols_published + ["P_PV_curtailment"]
    # Publish deferrable loads
    custom_deferrable_forecast_id = params["passed_data"][
        "custom_deferrable_forecast_id"
    ]
    for k in range(input_data_dict["opt"].optim_conf["num_def_loads"]):
        if "P_deferrable{}".format(k) not in opt_res_latest.columns:
            logger.error(
                "P_deferrable{}".format(k)
                + " was not found in results DataFrame. Optimization task may need to be relaunched or it did not converge to a solution.",
            )
        else:
            input_data_dict["rh"].post_data(
                opt_res_latest["P_deferrable{}".format(k)],
                idx_closest,
                custom_deferrable_forecast_id[k]["entity_id"],
                custom_deferrable_forecast_id[k]["unit_of_measurement"],
                custom_deferrable_forecast_id[k]["friendly_name"],
                type_var="deferrable",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post
            )
            cols_published = cols_published + ["P_deferrable{}".format(k)]
    # Publish battery power
    if input_data_dict["opt"].optim_conf["set_use_battery"]:
        if "P_batt" not in opt_res_latest.columns:
            logger.error(
                "P_batt was not found in results DataFrame. Optimization task may need to be relaunched or it did not converge to a solution.",
            )
        else:
            custom_batt_forecast_id = params["passed_data"]["custom_batt_forecast_id"]
            input_data_dict["rh"].post_data(
                opt_res_latest["P_batt"],
                idx_closest,
                custom_batt_forecast_id["entity_id"],
                custom_batt_forecast_id["unit_of_measurement"],
                custom_batt_forecast_id["friendly_name"],
                type_var="batt",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post
            )
            cols_published = cols_published + ["P_batt"]
            custom_batt_soc_forecast_id = params["passed_data"][
                "custom_batt_soc_forecast_id"
            ]
            input_data_dict["rh"].post_data(
                opt_res_latest["SOC_opt"] * 100,
                idx_closest,
                custom_batt_soc_forecast_id["entity_id"],
                custom_batt_soc_forecast_id["unit_of_measurement"],
                custom_batt_soc_forecast_id["friendly_name"],
                type_var="SOC",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post
            )
            cols_published = cols_published + ["SOC_opt"]
    # Publish grid power
    custom_grid_forecast_id = params["passed_data"]["custom_grid_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_grid"],
        idx_closest,
        custom_grid_forecast_id["entity_id"],
        custom_grid_forecast_id["unit_of_measurement"],
        custom_grid_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = cols_published + ["P_grid"]
    # Publish total value of cost function
    custom_cost_fun_id = params["passed_data"]["custom_cost_fun_id"]
    col_cost_fun = [i for i in opt_res_latest.columns if "cost_fun_" in i]
    input_data_dict["rh"].post_data(
        opt_res_latest[col_cost_fun],
        idx_closest,
        custom_cost_fun_id["entity_id"],
        custom_cost_fun_id["unit_of_measurement"],
        custom_cost_fun_id["friendly_name"],
        type_var="cost_fun",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    # cols_published = cols_published + col_cost_fun
    # Publish the optimization status
    custom_cost_fun_id = params["passed_data"]["custom_optim_status_id"]
    if "optim_status" not in opt_res_latest:
        opt_res_latest["optim_status"] = "Optimal"
        logger.warning(
            "no optim_status in opt_res_latest, run an optimization task first",
        )
    input_data_dict["rh"].post_data(
        opt_res_latest["optim_status"],
        idx_closest,
        custom_cost_fun_id["entity_id"],
        custom_cost_fun_id["unit_of_measurement"],
        custom_cost_fun_id["friendly_name"],
        type_var="optim_status",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = cols_published + ["optim_status"]
    # Publish unit_load_cost
    custom_unit_load_cost_id = params["passed_data"]["custom_unit_load_cost_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["unit_load_cost"],
        idx_closest,
        custom_unit_load_cost_id["entity_id"],
        custom_unit_load_cost_id["unit_of_measurement"],
        custom_unit_load_cost_id["friendly_name"],
        type_var="unit_load_cost",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = cols_published + ["unit_load_cost"]
    # Publish unit_prod_price
    custom_unit_prod_price_id = params["passed_data"]["custom_unit_prod_price_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["unit_prod_price"],
        idx_closest,
        custom_unit_prod_price_id["entity_id"],
        custom_unit_prod_price_id["unit_of_measurement"],
        custom_unit_prod_price_id["friendly_name"],
        type_var="unit_prod_price",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post
    )
    cols_published = cols_published + ["unit_prod_price"]
    # Create a DF resuming what has been published
    opt_res = opt_res_latest[cols_published].loc[[
        opt_res_latest.index[idx_closest]]]
    return opt_res

def continual_publish(input_data_dict: dict, entity_path: pathlib.Path, logger: logging.Logger):
    """
    If continual_publish is true and a entity file saved in /data_path/entities, continually publish sensor on freq rate, updating entity current state value based on timestamp

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param entity_path: Path for entities folder in data_path
    :type entity_path: Path
    :param logger: The passed logger object
    :type logger: logging.Logger

    """
    logger.info("Continual publish thread service started")
    freq = input_data_dict['retrieve_hass_conf'].get("freq", pd.to_timedelta(1, "minutes"))
    entity_path_contents = []
    while True:
        # Sleep for x seconds (using current time as a reference for time left)
        time.sleep(max(0,freq.total_seconds() - (datetime.now(input_data_dict["retrieve_hass_conf"]["time_zone"]).timestamp() % 60)))
        # Loop through all saved entity files
        if os.path.exists(entity_path) and len(os.listdir(entity_path)) > 0:
            entity_path_contents =  os.listdir(entity_path)    
            for entity in entity_path_contents:
                if entity != "metadata.json":
                    # Call publish_json with entity file, build entity, and publish                     
                    publish_json(entity, input_data_dict, entity_path, logger, "continual_publish")
        pass 
    # This function should never return           
    return False 

def publish_json(entity: dict, input_data_dict: dict, entity_path: pathlib.Path, 
                 logger: logging.Logger, reference: Optional[str] = ""):
    """
    Extract saved entity data from .json (in data_path/entities), build entity, post results to post_data

    :param entity: json file containing entity data
    :type entity: dict
    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param entity_path: Path for entities folder in data_path
    :type entity_path: Path
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param reference: String for identifying who ran the function  
    :type reference: str, optional
    
    """
    # Retrieve entity metadata from file
    if os.path.isfile(entity_path / "metadata.json"):
        with open(entity_path / "metadata.json", "r") as file:
            metadata = json.load(file) 
            if not metadata.get("lowest_freq",None) == None:
                freq = pd.to_timedelta(metadata["lowest_freq"], "minutes")
    else:
        logger.error("unable to located metadata.json in:" + entity_path)
        return False            
    # Round current timecode (now)
    now_precise = datetime.now(input_data_dict["retrieve_hass_conf"]["time_zone"]).replace(second=0, microsecond=0)
    # Retrieve entity data from file
    entity_data = pd.read_json(entity_path / entity , orient='index')
    # Remove ".json" from string for entity_id
    entity_id = entity.replace(".json", "")
    # Adjust Dataframe from received entity json file
    entity_data.columns = [metadata[entity_id]["name"]]
    entity_data.index.name = "timestamp"
    entity_data.index = pd.to_datetime(entity_data.index).tz_convert(input_data_dict["retrieve_hass_conf"]["time_zone"])
    entity_data.index.freq = pd.to_timedelta(int(metadata[entity_id]["freq"]), "minutes")    
    # Calculate the current state value
    if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
        idx_closest = entity_data.index.get_indexer([now_precise], method="nearest")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
        idx_closest = entity_data.index.get_indexer([now_precise], method="ffill")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
        idx_closest = entity_data.index.get_indexer([now_precise], method="bfill")[0]
    if idx_closest == -1:
        idx_closest = entity_data.index.get_indexer([now_precise], method="nearest")[0]
    # Call post data 
    if reference == "continual_publish":
        logger.debug("Auto Published sensor:")
        logger_levels = "DEBUG"
    else: 
        logger_levels = "INFO"
    # post/save entity
    input_data_dict["rh"].post_data(
        data_df=entity_data[metadata[entity_id]["name"]],
        idx=idx_closest,
        entity_id=entity_id,
        unit_of_measurement=metadata[entity_id]["unit_of_measurement"],
        friendly_name=metadata[entity_id]["friendly_name"],
        type_var=metadata[entity_id].get("type_var",""),
        save_entities=False,
        logger_levels=logger_levels
    )
    return entity_data[metadata[entity_id]["name"]]


def main():
    r"""Define the main command line entry function.

    This function may take several arguments as inputs. You can type `emhass --help` to see the list of options:

    - action: Set the desired action, options are: perfect-optim, dayahead-optim,
      naive-mpc-optim, publish-data, forecast-model-fit, forecast-model-predict, forecast-model-tune

    - config: Define path to the config.yaml file

    - costfun: Define the type of cost function, options are: profit, cost, self-consumption

    - log2file: Define if we should log to a file or not

    - params: Configuration parameters passed from data/options.json if using the add-on

    - runtimeparams: Pass runtime optimization parameters as dictionnary

    - debug: Use True for testing purposes

    """
    # Parsing arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--action', type=str, help='Set the desired action, options are: perfect-optim, dayahead-optim,\
        naive-mpc-optim, publish-data, forecast-model-fit, forecast-model-predict, forecast-model-tune')
    parser.add_argument('--config', type=str,
                        help='Define path to the config.yaml file')
    parser.add_argument('--data', type=str,
                        help='Define path to the Data files (.csv & .pkl)')
    parser.add_argument('--root', type=str, help='Define path emhass root')
    parser.add_argument('--costfun', type=str, default='profit',
                        help='Define the type of cost function, options are: profit, cost, self-consumption')
    parser.add_argument('--log2file', type=strtobool, default='False',
                        help='Define if we should log to a file or not')
    parser.add_argument('--params', type=str, default=None,
                        help='Configuration parameters passed from data/options.json')
    parser.add_argument('--runtimeparams', type=str, default=None,
                        help='Pass runtime optimization parameters as dictionnary')
    parser.add_argument('--debug', type=strtobool,
                        default='False', help='Use True for testing purposes')
    args = parser.parse_args()
    # The path to the configuration files
    if args.config is not None:
        config_path = pathlib.Path(args.config)
    else:
        config_path = pathlib.Path(
            str(utils.get_root(__file__, num_parent=2) / 'config_emhass.yaml'))
    if args.data is not None:
        data_path = pathlib.Path(args.data)
    else:
        data_path = (config_path.parent / 'data/')
    if args.root is not None:
        root_path = pathlib.Path(args.root)
    else:
        root_path = config_path.parent
    emhass_conf = {}
    emhass_conf['config_path'] = config_path
    emhass_conf['data_path'] = data_path
    emhass_conf['root_path'] = root_path
    # create logger
    logger, ch = utils.get_logger(
        __name__, emhass_conf, save_to_file=bool(args.log2file))
    logger.debug("config path: " + str(config_path))
    logger.debug("data path: " + str(data_path))
    logger.debug("root path: " + str(root_path))
    if not config_path.exists():
        logger.error(
            "Could not find config_emhass.yaml file in: " + str(config_path))
        logger.error("Try setting config file path with --config")
        return False
    if not os.path.isdir(data_path):
        logger.error("Could not find data foulder in: " + str(data_path))
        logger.error("Try setting data path with --data")
        return False
    if not os.path.isdir(root_path / 'src'):
        logger.error("Could not find emhass/src foulder in: " + str(root_path))
        logger.error("Try setting emhass root path with --root")
        return False
    # Additionnal argument
    try:
        parser.add_argument(
            "--version",
            action="version",
            version="%(prog)s " + version("emhass"),
        )
        args = parser.parse_args()
    except Exception:
        logger.info(
            "Version not found for emhass package. Or importlib exited with PackageNotFoundError.",
        )
    # Setup parameters
    input_data_dict = set_input_data_dict(emhass_conf,
                                          args.costfun, args.params, args.runtimeparams, args.action,
                                          logger, args.debug)
    # Perform selected action
    if args.action == "perfect-optim":
        opt_res = perfect_forecast_optim(
            input_data_dict, logger, debug=args.debug)
    elif args.action == "dayahead-optim":
        opt_res = dayahead_forecast_optim(
            input_data_dict, logger, debug=args.debug)
    elif args.action == "naive-mpc-optim":
        opt_res = naive_mpc_optim(input_data_dict, logger, debug=args.debug)
    elif args.action == "forecast-model-fit":
        df_fit_pred, df_fit_pred_backtest, mlf = forecast_model_fit(
            input_data_dict, logger, debug=args.debug
        )
        opt_res = None
    elif args.action == "forecast-model-predict":
        if args.debug:
            _, _, mlf = forecast_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlf = None
        df_pred = forecast_model_predict(input_data_dict, logger, debug=args.debug, mlf=mlf)
        opt_res = None
    elif args.action == "forecast-model-tune":
        if args.debug:
            _, _, mlf = forecast_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlf = None
        df_pred_optim, mlf = forecast_model_tune(input_data_dict, logger, debug=args.debug, mlf=mlf)
        opt_res = None
    elif args.action == "regressor-model-fit":
        mlr = regressor_model_fit(input_data_dict, logger, debug=args.debug)
        opt_res = None
    elif args.action == "regressor-model-predict":
        if args.debug:
            mlr = regressor_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlr = None
        prediction = regressor_model_predict(input_data_dict, logger, debug=args.debug,mlr=mlr)
        opt_res = None
    elif args.action == "publish-data":
        opt_res = publish_data(input_data_dict,logger)
    else:
        logger.error("The passed action argument is not valid")
        logger.error("Try setting --action: perfect-optim, dayahead-optim, naive-mpc-optim, forecast-model-fit, forecast-model-predict, forecast-model-tune or publish-data")
        opt_res = None
    logger.info(opt_res)
    # Flush the logger
    ch.close()
    logger.removeHandler(ch)
    if (
        args.action == "perfect-optim"
        or args.action == "dayahead-optim"
        or args.action == "naive-mpc-optim"
        or args.action == "publish-data"
    ):
        return opt_res
    elif args.action == "forecast-model-fit":
        return df_fit_pred, df_fit_pred_backtest, mlf
    elif args.action == "forecast-model-predict":
        return df_pred
    elif args.action == "regressor-model-fit":
        return mlr
    elif args.action == "regressor-model-predict":
        return prediction
    elif args.action == "forecast-model-tune":
        return df_pred_optim, mlf
    else:
        return opt_res


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
import os
import pickle
import copy
import logging
import json
from typing import Optional
import bz2
import pickle as cPickle
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from requests import get
from bs4 import BeautifulSoup
import pvlib
from pvlib.pvsystem import PVSystem
from pvlib.location import Location
from pvlib.modelchain import ModelChain
from pvlib.temperature import TEMPERATURE_MODEL_PARAMETERS
from pvlib.irradiance import disc

from emhass.retrieve_hass import RetrieveHass
from emhass.machine_learning_forecaster import MLForecaster
from emhass.utils import get_days_list, set_df_index_freq


class Forecast(object):
    r"""
    Generate weather, load and costs forecasts needed as inputs to the optimization.
    
    In EMHASS we have basically 4 forecasts to deal with:
    
    - PV power production forecast (internally based on the weather forecast and the
      characteristics of your PV plant). This is given in Watts.
    
    - Load power forecast: how much power your house will demand on the next 24h. This
      is given in Watts.
    
    - PV production selling price forecast: at what price are you selling your excess
      PV production on the next 24h. This is given in EUR/kWh.
    
    - Load cost forecast: the price of the energy from the grid on the next 24h. This
      is given in EUR/kWh.
    
    There are methods that are generalized to the 4 forecast needed. For all there
    forecasts it is possible to pass the data either as a passed list of values or by
    reading from a CSV file. With these methods it is then possible to use data from
    external forecast providers.
    
    Then there are the methods that are specific to each type of forecast and that 
    proposed forecast treated and generated internally by this EMHASS forecast class.
    For the weather forecast a first method (`scrapper`) uses a scrapping to the 
    ClearOutside webpage which proposes detailed forecasts based on Lat/Lon locations. 
    This method seems stable but as with any scrape method it will fail if any changes 
    are made to the webpage API. Another method (`solcast`) is using the SolCast PV 
    production forecast service. A final method (`solar.forecast`) is using another 
    external service: Solar.Forecast, for which just the nominal PV peak installed 
    power should be provided. Search the forecast section on the documentation for examples 
    on how to implement these different methods.
    
    The `get_power_from_weather` method is proposed here to convert from irradiance
    data to electrical power. The PVLib module is used to model the PV plant.
    
    The specific methods for the load forecast are a first method (`naive`) that uses 
    a naive approach, also called persistance. It simply assumes that the forecast for 
    a future period will be equal to the observed values in a past period. The past 
    period is controlled using parameter `delta_forecast`. A second method (`mlforecaster`)
    uses an internal custom forecasting model using machine learning. There is a section
    in the documentation explaining how to use this method.
    
    .. note:: This custom machine learning model is introduced from v0.4.0. EMHASS \
        proposed this new `mlforecaster` class with `fit`, `predict` and `tune` methods. \
        Only the `predict` method is used here to generate new forecasts, but it is \
        necessary to previously fit a forecaster model and it is a good idea to \
        optimize the model hyperparameters using the `tune` method. See the dedicated \
        section in the documentation for more help.
    
    For the PV production selling price and Load cost forecasts the privileged method
    is a direct read from a user provided list of values. The list should be passed
    as a runtime parameter during the `curl` to the EMHASS API.
    
    I reading from a CSV file, it should contain no header and the timestamped data 
    should have the following format:
    
    2021-04-29 00:00:00+00:00,287.07
    
    2021-04-29 00:30:00+00:00,274.27
    
    2021-04-29 01:00:00+00:00,243.38
    
    ...
    
    The data columns in these files will correspond to the data in the units expected
    for each forecasting method.
    
    """

    def __init__(self, retrieve_hass_conf: dict, optim_conf: dict, plant_conf: dict, 
                 params: str, emhass_conf: dict, logger: logging.Logger, 
                 opt_time_delta: Optional[int] = 24,
                 get_data_from_file: Optional[bool] = False) -> None:
        """
        Define constructor for the forecast class.
        
        :param retrieve_hass_conf: Dictionary containing the needed configuration
            data from the configuration file, specific to retrieve data from HASS
        :type retrieve_hass_conf: dict
        :param optim_conf: Dictionary containing the needed configuration
            data from the configuration file, specific for the optimization task
        :type optim_conf: dict
        :param plant_conf: Dictionary containing the needed configuration
            data from the configuration file, specific for the modeling of the PV plant
        :type plant_conf: dict
        :param params: Configuration parameters passed from data/options.json
        :type params: str
        :param emhass_conf: Dictionary containing the needed emhass paths
        :type emhass_conf: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param opt_time_delta: The time delta in hours used to generate forecasts, 
            a value of 24 will generate 24 hours of forecast data, defaults to 24
        :type opt_time_delta: int, optional
        :param get_data_from_file: Select if data should be retrieved from a 
            previously saved pickle useful for testing or directly from connection to
            hass database
        :type get_data_from_file: bool, optional

        """
        self.retrieve_hass_conf = retrieve_hass_conf
        self.optim_conf = optim_conf
        self.plant_conf = plant_conf
        self.freq = self.retrieve_hass_conf['freq']
        self.time_zone = self.retrieve_hass_conf['time_zone']
        self.method_ts_round = self.retrieve_hass_conf['method_ts_round']
        self.timeStep = self.freq.seconds/3600 # in hours
        self.time_delta = pd.to_timedelta(opt_time_delta, "hours")
        self.var_PV = self.retrieve_hass_conf['var_PV']
        self.var_load = self.retrieve_hass_conf['var_load']
        self.var_load_new = self.var_load+'_positive'
        self.lat = self.retrieve_hass_conf['lat'] 
        self.lon = self.retrieve_hass_conf['lon']
        self.emhass_conf = emhass_conf
        self.logger = logger
        self.get_data_from_file = get_data_from_file
        self.var_load_cost = 'unit_load_cost'
        self.var_prod_price = 'unit_prod_price'
        if params is None:
            self.params = params
        else:
            self.params = json.loads(params)
        if self.method_ts_round == 'nearest':
            self.start_forecast = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0)
        elif self.method_ts_round == 'first':
            self.start_forecast = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0).floor(freq=self.freq)
        elif self.method_ts_round == 'last':
            self.start_forecast = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0).ceil(freq=self.freq)
        else:
            self.logger.error("Wrong method_ts_round passed parameter")
        self.end_forecast = (self.start_forecast + self.optim_conf['delta_forecast']).replace(microsecond=0)
        self.forecast_dates = pd.date_range(start=self.start_forecast, 
                                            end=self.end_forecast-self.freq, 
                                            freq=self.freq).round(self.freq, ambiguous='infer', nonexistent='shift_forward')
        if params is not None:
            if 'prediction_horizon' in list(self.params['passed_data'].keys()):
                if self.params['passed_data']['prediction_horizon'] is not None:
                    self.forecast_dates = self.forecast_dates[0:self.params['passed_data']['prediction_horizon']]
        
        
    def get_weather_forecast(self, method: Optional[str] = 'scrapper',
                             csv_path: Optional[str] = "data_weather_forecast.csv") -> pd.DataFrame:
        r"""
        Get and generate weather forecast data.
        
        :param method: The desired method, options are 'scrapper', 'csv', 'list', 'solcast' and \
            'solar.forecast'. Defaults to 'scrapper'.
        :type method: str, optional
        :return: The DataFrame containing the forecasted data
        :rtype: pd.DataFrame
        
        """
        csv_path  = self.emhass_conf['data_path'] / csv_path

        self.logger.info("Retrieving weather forecast data using method = "+method)
        self.weather_forecast_method = method # Saving this attribute for later use to identify csv method usage
        if method == 'scrapper':
            freq_scrap = pd.to_timedelta(60, "minutes") # The scrapping time step is 60min on clearoutside
            forecast_dates_scrap = pd.date_range(start=self.start_forecast,
                                                 end=self.end_forecast-freq_scrap, 
                                                 freq=freq_scrap).round(freq_scrap, ambiguous='infer', nonexistent='shift_forward')
            # Using the clearoutside webpage
            response = get("https://clearoutside.com/forecast/"+str(round(self.lat, 2))+"/"+str(round(self.lon, 2))+"?desktop=true")
            '''import bz2 # Uncomment to save a serialized data for tests
            import _pickle as cPickle
            with bz2.BZ2File("data/test_response_scrapper_get_method.pbz2", "w") as f: 
                cPickle.dump(response.content, f)'''
            soup = BeautifulSoup(response.content, 'html.parser')
            table = soup.find_all(id='day_0')[0]
            list_names = table.find_all(class_='fc_detail_label')
            list_tables = table.find_all('ul')[1:]
            selected_cols = [0, 1, 2, 3, 10, 12, 15] # Selected variables
            col_names = [list_names[i].get_text() for i in selected_cols]
            list_tables = [list_tables[i] for i in selected_cols]
            # Building the raw DF container
            raw_data = pd.DataFrame(index=range(len(forecast_dates_scrap)), columns=col_names, dtype=float)
            for count_col, col in enumerate(col_names):
                list_rows = list_tables[count_col].find_all('li')
                for count_row, row in enumerate(list_rows):
                    raw_data.loc[count_row, col] = float(row.get_text())
            # Treating index
            raw_data.set_index(forecast_dates_scrap, inplace=True)
            raw_data = raw_data[~raw_data.index.duplicated(keep='first')]
            raw_data = raw_data.reindex(self.forecast_dates)
            raw_data.interpolate(method='linear', axis=0, limit=None, 
                                 limit_direction='both', inplace=True)
            # Converting the cloud cover into Global Horizontal Irradiance with a PVLib method
            ghi_est = self.cloud_cover_to_irradiance(raw_data['Total Clouds (% Sky Obscured)'])
            data = ghi_est
            data['temp_air'] = raw_data['Temperature (°C)']
            data['wind_speed'] = raw_data['Wind Speed/Direction (mph)']*1.60934 # conversion to km/h
            data['relative_humidity'] = raw_data['Relative Humidity (%)']
            data['precipitable_water'] = pvlib.atmosphere.gueymard94_pw(
                data['temp_air'], data['relative_humidity'])
        elif method == 'solcast': # using solcast API
            # Retrieve data from the solcast API
            if 'solcast_api_key' not in self.retrieve_hass_conf:
                self.logger.warning("The solcast_api_key parameter was not defined, using dummy values for testing")
                self.retrieve_hass_conf['solcast_api_key'] = "123456"
            if 'solcast_rooftop_id' not in self.retrieve_hass_conf:
                self.logger.warning("The solcast_rooftop_id parameter was not defined, using dummy values for testing")
                self.retrieve_hass_conf['solcast_rooftop_id'] = "123456"
            headers = {
                "Authorization": "Bearer " + self.retrieve_hass_conf['solcast_api_key'],
                "content-type": "application/json",
                }
            days_solcast = int(len(self.forecast_dates)*self.freq.seconds/3600)
            url = "https://api.solcast.com.au/rooftop_sites/"+self.retrieve_hass_conf['solcast_rooftop_id']+"/forecasts?hours="+str(days_solcast)
            response = get(url, headers=headers)
            '''import bz2 # Uncomment to save a serialized data for tests
            import _pickle as cPickle
            with bz2.BZ2File("data/test_response_solcast_get_method.pbz2", "w") as f: 
                cPickle.dump(response, f)'''
            data = response.json()
            data_list = []
            for elm in data['forecasts']:
                data_list.append(elm['pv_estimate']*1000) # Converting kW to W
            # Check if the retrieved data has the correct length
            if len(data_list) < len(self.forecast_dates):
                self.logger.error("Not enough data retrived from SolCast service, try increasing the time step or use MPC")
            else:
                # Ensure correct length
                data_list = data_list[0:len(self.forecast_dates)]
                # Define DataFrame
                data_dict = {'ts':self.forecast_dates, 'yhat':data_list}
                data = pd.DataFrame.from_dict(data_dict)
                # Define index
                data.set_index('ts', inplace=True)
        elif method == 'solar.forecast': # using the solar.forecast API
            # Retrieve data from the solar.forecast API
            if 'solar_forecast_kwp' not in self.retrieve_hass_conf:
                self.logger.warning("The solar_forecast_kwp parameter was not defined, using dummy values for testing")
                self.retrieve_hass_conf['solar_forecast_kwp'] = 5
            if  self.retrieve_hass_conf['solar_forecast_kwp'] == 0:
                self.logger.warning("The solar_forecast_kwp parameter is set to zero, setting to default 5")
                self.retrieve_hass_conf['solar_forecast_kwp'] = 5
            if self.optim_conf['delta_forecast'].days > 1:
                self.logger.warning("The free public tier for solar.forecast only provides one day forecasts")
                self.logger.warning("Continuing with just the first day of data, the other days are filled with 0.0.")
                self.logger.warning("Use the other available methods for delta_forecast > 1")
            headers = {
                "Accept": "application/json"
                }
            data = pd.DataFrame()
            for i in range(len(self.plant_conf['module_model'])):
                url = "https://api.forecast.solar/estimate/"+str(round(self.lat, 2))+"/"+str(round(self.lon, 2))+\
                    "/"+str(self.plant_conf["surface_tilt"][i])+"/"+str(self.plant_conf["surface_azimuth"][i]-180)+\
                    "/"+str(self.retrieve_hass_conf["solar_forecast_kwp"])
                response = get(url, headers=headers)
                '''import bz2 # Uncomment to save a serialized data for tests
                import _pickle as cPickle
                with bz2.BZ2File("data/test_response_solarforecast_get_method.pbz2", "w") as f: 
                    cPickle.dump(response.json(), f)'''
                data_raw = response.json()
                data_dict = {'ts':list(data_raw['result']['watts'].keys()), 'yhat':list(data_raw['result']['watts'].values())}
                # Form the final DataFrame
                data_tmp = pd.DataFrame.from_dict(data_dict)
                data_tmp.set_index('ts', inplace=True)
                data_tmp.index = pd.to_datetime(data_tmp.index)
                data_tmp = data_tmp.tz_localize(self.forecast_dates.tz)
                data_tmp = data_tmp.reindex(index=self.forecast_dates)
                mask_up_data_df = data_tmp.copy(deep=True).fillna(method = "ffill").isnull()
                mask_down_data_df = data_tmp.copy(deep=True).fillna(method = "bfill").isnull()
                data_tmp.loc[data_tmp.index[mask_up_data_df['yhat']==True],:] = 0.0
                data_tmp.loc[data_tmp.index[mask_down_data_df['yhat']==True],:] = 0.0
                data_tmp.interpolate(inplace=True, limit=1)
                data_tmp = data_tmp.fillna(0.0)
                if len(data) == 0:
                    data = copy.deepcopy(data_tmp)
                else:
                    data = data + data_tmp
        elif method == 'csv': # reading from a csv file
            weather_csv_file_path = csv_path
            # Loading the csv file, we will consider that this is the PV power in W
            data = pd.read_csv(weather_csv_file_path, header=None, names=['ts', 'yhat'])
            # Check if the passed data has the correct length       
            if len(data) < len(self.forecast_dates):
                self.logger.error("Passed data from CSV is not long enough")
            else:
                # Ensure correct length
                data = data.loc[data.index[0:len(self.forecast_dates)],:]
                # Define index
                data.index = self.forecast_dates
                data.drop('ts', axis=1, inplace=True)
                data = data.copy().loc[self.forecast_dates]
        elif method == 'list': # reading a list of values
            # Loading data from passed list
            data_list = self.params['passed_data']['pv_power_forecast']
            # Check if the passed data has the correct length
            if len(data_list) < len(self.forecast_dates) and self.params['passed_data']['prediction_horizon'] is None:
                self.logger.error("Passed data from passed list is not long enough")
            else:
                # Ensure correct length
                data_list = data_list[0:len(self.forecast_dates)]
                # Define DataFrame
                data_dict = {'ts':self.forecast_dates, 'yhat':data_list}
                data = pd.DataFrame.from_dict(data_dict)
                # Define index
                data.set_index('ts', inplace=True)
        else:
            self.logger.error("Method %r is not valid", method)
            data = None
        return data
    
    def cloud_cover_to_irradiance(self, cloud_cover: pd.Series, 
                                  offset:Optional[int] = 35) -> pd.DataFrame:
        """
        Estimates irradiance from cloud cover in the following steps.
        
        1. Determine clear sky GHI using Ineichen model and
           climatological turbidity.
           
        2. Estimate cloudy sky GHI using a function of cloud_cover
           
        3. Estimate cloudy sky DNI using the DISC model.
        
        4. Calculate DHI from DNI and GHI.
        
        (This function was copied and modified from PVLib)

        :param cloud_cover: Cloud cover in %.
        :type cloud_cover: pd.Series
        :param offset: Determines the minimum GHI., defaults to 35
        :type offset: Optional[int], optional
        :return: Estimated GHI, DNI, and DHI.
        :rtype: pd.DataFrame
        """
        location = Location(latitude=self.lat, longitude=self.lon)
        solpos = location.get_solarposition(cloud_cover.index)
        cs = location.get_clearsky(cloud_cover.index, model='ineichen', 
                                   solar_position=solpos)
        # Using only the linear method
        offset = offset / 100.
        cloud_cover_unit = copy.deepcopy(cloud_cover) / 100.
        ghi = (offset + (1 - offset) * (1 - cloud_cover_unit)) * cs['ghi']
        # Using disc model
        dni = disc(ghi, solpos['zenith'], cloud_cover.index)['dni']
        dhi = ghi - dni * np.cos(np.radians(solpos['zenith']))
        irrads = pd.DataFrame({'ghi': ghi, 'dni': dni, 'dhi': dhi}).fillna(0)
        return irrads
    
    @staticmethod
    def get_mix_forecast(df_now: pd.DataFrame, df_forecast: pd.DataFrame, 
                         alpha:float, beta:float, col:str) -> pd.DataFrame:
        """A simple correction method for forecasted data using the current real values of a variable.

        :param df_now: The DataFrame containing the current/real values
        :type df_now: pd.DataFrame
        :param df_forecast: The DataFrame containing the forecast data
        :type df_forecast: pd.DataFrame
        :param alpha: A weight for the forecast data side
        :type alpha: float
        :param beta: A weight for the current/real values sied
        :type beta: float
        :param col: The column variable name
        :type col: str
        :return: The output DataFrame with the corrected values
        :rtype: pd.DataFrame
        """
        first_fcst = alpha*df_forecast.iloc[0] + beta*df_now[col].iloc[-1]
        df_forecast.iloc[0] = first_fcst
        return df_forecast
    
    def get_power_from_weather(self, df_weather: pd.DataFrame, 
                               set_mix_forecast:Optional[bool] = False,
                               df_now:Optional[pd.DataFrame] = pd.DataFrame()) -> pd.Series:
        r"""
        Convert wheater forecast data into electrical power.
        
        :param df_weather: The DataFrame containing the weather forecasted data. \
            This DF should be generated by the 'get_weather_forecast' method or at \
            least contain the same columns names filled with proper data.
        :type df_weather: pd.DataFrame
        :param set_mix_forecast: Use a mixed forcast strategy to integra now/current values.
        :type set_mix_forecast: Bool, optional
        :param df_now: The DataFrame containing the now/current data.
        :type df_now: pd.DataFrame
        :return: The DataFrame containing the electrical power in Watts
        :rtype: pd.DataFrame

        """
        # If using csv method we consider that yhat is the PV power in W
        if "solar_forecast_kwp" in self.retrieve_hass_conf.keys() and self.retrieve_hass_conf["solar_forecast_kwp"] == 0:
            P_PV_forecast = pd.Series(0, index=df_weather.index)
        else:
            if self.weather_forecast_method == 'solcast' or self.weather_forecast_method == 'solar.forecast' or \
                self.weather_forecast_method == 'csv' or self.weather_forecast_method == 'list':
                P_PV_forecast = df_weather['yhat']
                P_PV_forecast.name = None
            else: # We will transform the weather data into electrical power
                # Transform to power (Watts)
                # Setting the main parameters of the PV plant
                location = Location(latitude=self.lat, longitude=self.lon)
                temp_params = TEMPERATURE_MODEL_PARAMETERS['sapm']['close_mount_glass_glass']
                cec_modules = bz2.BZ2File(self.emhass_conf['root_path'] / 'data/cec_modules.pbz2', "rb")
                cec_modules = cPickle.load(cec_modules)
                cec_inverters = bz2.BZ2File(self.emhass_conf['root_path'] / 'data/cec_inverters.pbz2', "rb")
                cec_inverters = cPickle.load(cec_inverters)
                if type(self.plant_conf['module_model']) == list:
                    P_PV_forecast = pd.Series(0, index=df_weather.index)
                    for i in range(len(self.plant_conf['module_model'])):
                        # Selecting correct module and inverter
                        module = cec_modules[self.plant_conf['module_model'][i]]
                        inverter = cec_inverters[self.plant_conf['inverter_model'][i]]
                        # Building the PV system in PVLib
                        system = PVSystem(surface_tilt=self.plant_conf['surface_tilt'][i], 
                                        surface_azimuth=self.plant_conf['surface_azimuth'][i],
                                        module_parameters=module,
                                        inverter_parameters=inverter,
                                        temperature_model_parameters=temp_params,
                                        modules_per_string=self.plant_conf['modules_per_string'][i],
                                        strings_per_inverter=self.plant_conf['strings_per_inverter'][i])
                        mc = ModelChain(system, location, aoi_model="physical")
                        # Run the model on the weather DF indexes
                        mc.run_model(df_weather)
                        # Extracting results for AC power
                        P_PV_forecast = P_PV_forecast + mc.results.ac
                else:
                    # Selecting correct module and inverter
                    module = cec_modules[self.plant_conf['module_model']]
                    inverter = cec_inverters[self.plant_conf['inverter_model']]
                    # Building the PV system in PVLib
                    system = PVSystem(surface_tilt=self.plant_conf['surface_tilt'], 
                                    surface_azimuth=self.plant_conf['surface_azimuth'],
                                    module_parameters=module,
                                    inverter_parameters=inverter,
                                    temperature_model_parameters=temp_params,
                                    modules_per_string=self.plant_conf['modules_per_string'],
                                    strings_per_inverter=self.plant_conf['strings_per_inverter'])
                    mc = ModelChain(system, location, aoi_model="physical")
                    # Run the model on the weather DF indexes
                    mc.run_model(df_weather)
                    # Extracting results for AC power
                    P_PV_forecast = mc.results.ac
        if set_mix_forecast:
            P_PV_forecast = Forecast.get_mix_forecast(
                df_now, P_PV_forecast, 
                self.params['passed_data']['alpha'], self.params['passed_data']['beta'], self.var_PV)
        return P_PV_forecast
    
    def get_forecast_days_csv(self, timedelta_days: Optional[int] = 1) -> pd.date_range:
        r"""
        Get the date range vector of forecast dates that will be used when loading a CSV file.
        
        :return: The forecast dates vector
        :rtype: pd.date_range

        """
        start_forecast_csv = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0)
        if self.method_ts_round == 'nearest':
            start_forecast_csv = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0)
        elif self.method_ts_round == 'first':
            start_forecast_csv = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0).floor(freq=self.freq)
        elif self.method_ts_round == 'last':
            start_forecast_csv = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(microsecond=0).ceil(freq=self.freq)
        else:
            self.logger.error("Wrong method_ts_round passed parameter")
        end_forecast_csv = (start_forecast_csv + self.optim_conf['delta_forecast']).replace(microsecond=0)
        forecast_dates_csv = pd.date_range(start=start_forecast_csv, 
                                           end=end_forecast_csv+timedelta(days=timedelta_days)-self.freq, 
                                           freq=self.freq).round(self.freq, ambiguous='infer', nonexistent='shift_forward')
        if self.params is not None:
            if 'prediction_horizon' in list(self.params['passed_data'].keys()):
                if self.params['passed_data']['prediction_horizon'] is not None:
                    forecast_dates_csv = forecast_dates_csv[0:self.params['passed_data']['prediction_horizon']]
        return forecast_dates_csv
    
    def get_forecast_out_from_csv_or_list(self, df_final: pd.DataFrame, forecast_dates_csv: pd.date_range,
                                          csv_path: str, data_list: Optional[list] = None, 
                                          list_and_perfect: Optional[bool] = False) -> pd.DataFrame:
        r"""
        Get the forecast data as a DataFrame from a CSV file. 
        
        The data contained in the CSV file should be a 24h forecast with the same frequency as 
        the main 'freq' parameter in the configuration file. The timestamp will not be used and 
        a new DateTimeIndex is generated to fit the timestamp index of the input data in 'df_final'.
        
        :param df_final: The DataFrame containing the input data.
        :type df_final: pd.DataFrame
        :param forecast_dates_csv: The forecast dates vector
        :type forecast_dates_csv: pd.date_range
        :param csv_path: The path to the CSV file
        :type csv_path: str
        :return: The data from the CSV file
        :rtype: pd.DataFrame

        """
        if csv_path is None:
            data_dict = {'ts':forecast_dates_csv, 'yhat':data_list}
            df_csv = pd.DataFrame.from_dict(data_dict)
            df_csv.index = forecast_dates_csv
            df_csv.drop(['ts'], axis=1, inplace=True)
            df_csv = set_df_index_freq(df_csv)
            if list_and_perfect:
                days_list = df_final.index.day.unique().tolist()
            else:
                days_list = df_csv.index.day.unique().tolist()
        else:
            if not os.path.exists(csv_path):
                csv_path = self.emhass_conf['data_path'] / csv_path
            load_csv_file_path = csv_path    
            df_csv = pd.read_csv(load_csv_file_path, header=None, names=['ts', 'yhat'])
            df_csv.index = forecast_dates_csv
            df_csv.drop(['ts'], axis=1, inplace=True)
            df_csv = set_df_index_freq(df_csv)
            days_list = df_final.index.day.unique().tolist()
        forecast_out = pd.DataFrame()
        for day in days_list:
            if csv_path is None:
                if list_and_perfect:
                    df_tmp = copy.deepcopy(df_final)
                else:
                    df_tmp = copy.deepcopy(df_csv)
            else:
                df_tmp = copy.deepcopy(df_final)
            first_elm_index = [i for i, x in enumerate(df_tmp.index.day == day) if x][0]
            last_elm_index = [i for i, x in enumerate(df_tmp.index.day == day) if x][-1]
            fcst_index = pd.date_range(start=df_tmp.index[first_elm_index],
                                    end=df_tmp.index[last_elm_index], 
                                    freq=df_tmp.index.freq)
            first_hour = str(df_tmp.index[first_elm_index].hour)+":"+str(df_tmp.index[first_elm_index].minute)
            last_hour = str(df_tmp.index[last_elm_index].hour)+":"+str(df_tmp.index[last_elm_index].minute)
            if len(forecast_out) == 0:
                if csv_path is None:
                    if list_and_perfect:
                        forecast_out = pd.DataFrame(
                            df_csv.between_time(first_hour, last_hour).values,
                            index=fcst_index)
                    else:
                        forecast_out = pd.DataFrame(
                            df_csv.loc[fcst_index,:].between_time(first_hour, last_hour).values,
                            index=fcst_index)
                else:
                    forecast_out = pd.DataFrame(
                        df_csv.between_time(first_hour, last_hour).values,
                        index=fcst_index)
            else:
                if csv_path is None:
                    if list_and_perfect:
                        forecast_tp = pd.DataFrame(
                            df_csv.between_time(first_hour, last_hour).values,
                            index=fcst_index)
                    else:
                        forecast_tp = pd.DataFrame(
                            df_csv.loc[fcst_index,:].between_time(first_hour, last_hour).values,
                            index=fcst_index)
                else:
                    forecast_tp = pd.DataFrame(
                        df_csv.between_time(first_hour, last_hour).values,
                        index=fcst_index)
                forecast_out = pd.concat([forecast_out, forecast_tp], axis=0)
        return forecast_out
    
    def get_load_forecast(self, days_min_load_forecast: Optional[int] = 3, method: Optional[str] = 'naive',
                          csv_path: Optional[str] = "data_load_forecast.csv",
                          set_mix_forecast:Optional[bool] = False, df_now:Optional[pd.DataFrame] = pd.DataFrame(),
                          use_last_window: Optional[bool] = True, mlf: Optional[MLForecaster] = None,
                          debug: Optional[bool] = False) -> pd.Series:
        r"""
        Get and generate the load forecast data.
        
        :param days_min_load_forecast: The number of last days to retrieve that \
            will be used to generate a naive forecast, defaults to 3
        :type days_min_load_forecast: int, optional
        :param method: The method to be used to generate load forecast, the options \
            are 'naive' for a persistance model, 'mlforecaster' for using a custom \
            previously fitted machine learning model, 'csv' to read the forecast from \
            a CSV file and 'list' to use data directly passed at runtime as a list of \
            values. Defaults to 'naive'.
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "/data/data_load_forecast.csv"
        :type csv_path: str, optional
        :param set_mix_forecast: Use a mixed forcast strategy to integra now/current values.
        :type set_mix_forecast: Bool, optional
        :param df_now: The DataFrame containing the now/current data.
        :type df_now: pd.DataFrame, optional
        :param use_last_window: True if the 'last_window' option should be used for the \
            custom machine learning forecast model. The 'last_window=True' means that the data \
            that will be used to generate the new forecast will be freshly retrieved from \
            Home Assistant. This data is needed because the forecast model is an auto-regressive \
            model with lags. If 'False' then the data using during the model train is used.
        :type use_last_window: Bool, optional
        :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
            and unit testing. In production the actual model will be read from a saved pickle file.
        :type mlf: mlforecaster, optional
        :param debug: The DataFrame containing the now/current data.
        :type debug: Bool, optional
        :return: The DataFrame containing the electrical load power in Watts
        :rtype: pd.DataFrame

        """
        csv_path  = self.emhass_conf['data_path'] / csv_path
        
        if method == 'naive' or method == 'mlforecaster': # retrieving needed data for these methods
            self.logger.info("Retrieving data from hass for load forecast using method = "+method)
            var_list = [self.var_load]
            var_replace_zero = None
            var_interp = [self.var_load]
            time_zone_load_foreacast = None
            # We will need to retrieve a new set of load data according to the days_min_load_forecast parameter
            rh = RetrieveHass(self.retrieve_hass_conf['hass_url'], self.retrieve_hass_conf['long_lived_token'], 
                               self.freq, time_zone_load_foreacast, self.params, self.emhass_conf, self.logger)
            if self.get_data_from_file:
                filename_path = self.emhass_conf['data_path'] / 'test_df_final.pkl'
                with open(filename_path, 'rb') as inp:
                    rh.df_final, days_list, var_list = pickle.load(inp)
                    self.var_load = var_list[0]
                    self.retrieve_hass_conf['var_load'] = self.var_load
                    var_interp = [var_list[0]]
                    self.var_list = [var_list[0]]
                    self.var_load_new = self.var_load+'_positive'
            else:
                days_list = get_days_list(days_min_load_forecast) 
                if not rh.get_data(days_list, var_list):
                    return False
            if  not rh.prepare_data(
                self.retrieve_hass_conf['var_load'], load_negative = self.retrieve_hass_conf['load_negative'],
                set_zero_min = self.retrieve_hass_conf['set_zero_min'], 
                var_replace_zero = var_replace_zero, var_interp = var_interp):
                return False
            df = rh.df_final.copy()[[self.var_load_new]]
        if method == 'naive': # using a naive approach
            mask_forecast_out = (df.index > days_list[-1] - self.optim_conf['delta_forecast'])
            forecast_out = df.copy().loc[mask_forecast_out]
            forecast_out = forecast_out.rename(columns={self.var_load_new: 'yhat'})
            # Force forecast_out length to avoid mismatches
            forecast_out = forecast_out.iloc[0:len(self.forecast_dates)]
            forecast_out.index = self.forecast_dates
        elif method == 'mlforecaster': # using a custom forecast model with machine learning
            # Load model
            model_type = self.params['passed_data']['model_type']
            filename = model_type+'_mlf.pkl'
            filename_path = self.emhass_conf['data_path'] / filename
            if not debug:
                if filename_path.is_file():
                    with open(filename_path, 'rb') as inp:
                        mlf = pickle.load(inp)
                else:
                    self.logger.error("The ML forecaster file was not found, please run a model fit method before this predict method")
                    return False
            # Make predictions
            if use_last_window:
                data_last_window = copy.deepcopy(df)
                data_last_window = data_last_window.rename(columns={self.var_load_new: self.var_load})
            else:
                data_last_window = None
            forecast_out = mlf.predict(data_last_window)
            # Force forecast length to avoid mismatches
            self.logger.debug("Number of ML predict forcast data generated (lags_opt): " + str(len(forecast_out.index)))
            self.logger.debug("Number of forcast dates obtained: " + str(len(self.forecast_dates)))
            if len(self.forecast_dates) < len(forecast_out.index):
                forecast_out = forecast_out.iloc[0:len(self.forecast_dates)]
            # To be removed once bug is fixed
            elif len(self.forecast_dates) > len(forecast_out.index):
                self.logger.error("Unable to obtain: " + str(len(self.forecast_dates))  + " lags_opt values from sensor: power load no var loads, check optimization_time_step/freq and historic_days_to_retrieve/days_to_retrieve parameters")
                return False
            # Define DataFrame
            data_dict = {'ts':self.forecast_dates, 'yhat':forecast_out.values.tolist()}
            data = pd.DataFrame.from_dict(data_dict)
            # Define index
            data.set_index('ts', inplace=True)
            forecast_out = data.copy().loc[self.forecast_dates]
        elif method == 'csv': # reading from a csv file
            load_csv_file_path = csv_path
            df_csv = pd.read_csv(load_csv_file_path, header=None, names=['ts', 'yhat'])
            if len(df_csv) < len(self.forecast_dates):
                self.logger.error("Passed data from CSV is not long enough")
            else:
                # Ensure correct length
                df_csv = df_csv.loc[df_csv.index[0:len(self.forecast_dates)],:]
                # Define index
                df_csv.index = self.forecast_dates
                df_csv.drop(['ts'], axis=1, inplace=True)
                forecast_out = df_csv.copy().loc[self.forecast_dates]
        elif method == 'list': # reading a list of values
            # Loading data from passed list
            data_list = self.params['passed_data']['load_power_forecast']
            # Check if the passed data has the correct length
            if len(data_list) < len(self.forecast_dates) and self.params['passed_data']['prediction_horizon'] is None:
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0:len(self.forecast_dates)]
                # Define DataFrame
                data_dict = {'ts':self.forecast_dates, 'yhat':data_list}
                data = pd.DataFrame.from_dict(data_dict)
                # Define index
                data.set_index('ts', inplace=True)
                forecast_out = data.copy().loc[self.forecast_dates]
        else:
            self.logger.error("Passed method is not valid")
            return False
        P_Load_forecast = copy.deepcopy(forecast_out['yhat'])
        if set_mix_forecast:
            P_Load_forecast = Forecast.get_mix_forecast(
                df_now, P_Load_forecast, 
                self.params['passed_data']['alpha'], self.params['passed_data']['beta'], self.var_load_new)
        return P_Load_forecast
    
    def get_load_cost_forecast(self, df_final: pd.DataFrame, method: Optional[str] = 'hp_hc_periods',
                               csv_path: Optional[str] = "data_load_cost_forecast.csv",
                               list_and_perfect: Optional[bool] = False) -> pd.DataFrame:
        r"""
        Get the unit cost for the load consumption based on multiple tariff \
        periods. This is the cost of the energy from the utility in a vector \
        sampled at the fixed freq value.
        
        :param df_final: The DataFrame containing the input data.
        :type df_final: pd.DataFrame
        :param method: The method to be used to generate load cost forecast, \
            the options are 'hp_hc_periods' for peak and non-peak hours contracts\
            and 'csv' to load a CSV file, defaults to 'hp_hc_periods'
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "data_load_cost_forecast.csv"
        :type csv_path: str, optional
        :return: The input DataFrame with one additionnal column appended containing
            the load cost for each time observation.
        :rtype: pd.DataFrame

        """
        csv_path  = self.emhass_conf['data_path'] / csv_path
        if method == 'hp_hc_periods':
            df_final[self.var_load_cost] = self.optim_conf['load_cost_hc']
            list_df_hp = []
            for key, period_hp in self.optim_conf['list_hp_periods'].items():
                list_df_hp.append(df_final[self.var_load_cost].between_time(
                    period_hp[0]['start'], period_hp[1]['end']))
            for df_hp in list_df_hp:
                df_final.loc[df_hp.index, self.var_load_cost] = self.optim_conf['load_cost_hp']
        elif method == 'csv':
            forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
            forecast_out = self.get_forecast_out_from_csv_or_list(
                df_final, forecast_dates_csv, csv_path)
            df_final[self.var_load_cost] = forecast_out
        elif method == 'list': # reading a list of values
            # Loading data from passed list
            data_list = self.params['passed_data']['load_cost_forecast']
            # Check if the passed data has the correct length
            if len(data_list) < len(self.forecast_dates) and self.params['passed_data']['prediction_horizon'] is None:
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0:len(self.forecast_dates)]
                # Define the correct dates
                forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
                forecast_out = self.get_forecast_out_from_csv_or_list(
                    df_final, forecast_dates_csv, None, data_list=data_list, list_and_perfect=list_and_perfect)
                # Fill the final DF
                df_final[self.var_load_cost] = forecast_out
        else:
            self.logger.error("Passed method is not valid")
            return False
        return df_final
    
    def get_prod_price_forecast(self, df_final: pd.DataFrame, method: Optional[str] = 'constant',
                                csv_path: Optional[str] = "data_prod_price_forecast.csv", 
                                list_and_perfect: Optional[bool] = False) -> pd.DataFrame:

        r"""
        Get the unit power production price for the energy injected to the grid.\
        This is the price of the energy injected to the utility in a vector \
        sampled at the fixed freq value.
        
        :param df_input_data: The DataFrame containing all the input data retrieved
            from hass
        :type df_input_data: pd.DataFrame
        :param method: The method to be used to generate the production price forecast, \
            the options are 'constant' for a fixed constant value and 'csv'\
            to load a CSV file, defaults to 'constant'
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "/data/data_load_cost_forecast.csv"
        :type csv_path: str, optional
        :return: The input DataFrame with one additionnal column appended containing
            the power production price for each time observation.
        :rtype: pd.DataFrame

        """
        csv_path  = self.emhass_conf['data_path'] / csv_path
        if method == 'constant':
            df_final[self.var_prod_price] = self.optim_conf['prod_sell_price']
        elif method == 'csv':
            forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
            forecast_out = self.get_forecast_out_from_csv_or_list(
                df_final, forecast_dates_csv, csv_path)
            df_final[self.var_prod_price] = forecast_out
        elif method == 'list': # reading a list of values
            # Loading data from passed list
            data_list = self.params['passed_data']['prod_price_forecast']
            # Check if the passed data has the correct length
            if len(data_list) < len(self.forecast_dates) and self.params['passed_data']['prediction_horizon'] is None:
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0:len(self.forecast_dates)]
                # Define the correct dates
                forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
                forecast_out = self.get_forecast_out_from_csv_or_list(
                    df_final, forecast_dates_csv, None, data_list=data_list, list_and_perfect=list_and_perfect)
                # Fill the final DF
                df_final[self.var_prod_price] = forecast_out
        else:
            self.logger.error("Passed method is not valid")
            return False
        return df_final
    



"""Machine learning regressor module."""

from __future__ import annotations

import copy
import time
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from sklearn.ensemble import (
    AdaBoostRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)
from sklearn.linear_model import Lasso, LinearRegression, Ridge
from sklearn.metrics import r2_score
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

if TYPE_CHECKING:
    import logging

warnings.filterwarnings("ignore", category=DeprecationWarning)

REGRESSION_METHODS = {
    "LinearRegression": {
        "model": LinearRegression(),
        "param_grid": {
            "linearregression__fit_intercept": [True, False],
            "linearregression__positive": [True, False],
        },
    },
    "RidgeRegression": {
        "model": Ridge(),
        "param_grid": {"ridge__alpha": [0.1, 1.0, 10.0]},
    },
    "LassoRegression": {
        "model": Lasso(),
        "param_grid": {"lasso__alpha": [0.1, 1.0, 10.0]},
    },
    "RandomForestRegression": {
        "model": RandomForestRegressor(),
        "param_grid": {"randomforestregressor__n_estimators": [50, 100, 200]},
    },
    "GradientBoostingRegression": {
        "model": GradientBoostingRegressor(),
        "param_grid": {
            "gradientboostingregressor__n_estimators": [50, 100, 200],
            "gradientboostingregressor__learning_rate": [0.01, 0.1, 0.2],
        },
    },
    "AdaBoostRegression": {
        "model": AdaBoostRegressor(),
        "param_grid": {
            "adaboostregressor__n_estimators": [50, 100, 200],
            "adaboostregressor__learning_rate": [0.01, 0.1, 0.2],
        },
    },
}


class MLRegressor:
    r"""A forecaster class using machine learning models.

    This class uses the `sklearn` module and the machine learning models are \
        from `scikit-learn`.

    It exposes two main methods:

    - `fit`: to train a model with the passed data.

    - `predict`: to obtain a forecast from a pre-trained model.

    """

    def __init__(self: MLRegressor, data: pd.DataFrame, model_type: str, regression_model: str,
                 features: list, target: str, timestamp: str, logger: logging.Logger) -> None:
        r"""Define constructor for the forecast class.

        :param data: The data that will be used for train/test
        :type data: pd.DataFrame
        :param model_type: A unique name defining this model and useful to identify \
            for what it will be used for.
        :type model_type: str
        :param regression_model: The model that will be used. For now only \
            this options are possible: `LinearRegression`, `RidgeRegression`, \
            `LassoRegression`, `RandomForestRegression`, \
            `GradientBoostingRegression` and `AdaBoostRegression`.
        :type regression_model: str
        :param features: A list of features. \
            Example: [`solar_production`, `degree_days`].
        :type features: list
        :param target: The target(to be predicted). \
            Example: `heating_hours`.
        :type target: str
        :param timestamp: If defined, the column key that has to be used of timestamp.
        :type timestamp: str
        :param logger: The passed logger object
        :type logger: logging.Logger
        """
        self.data = data
        self.features = features
        self.target = target
        self.timestamp = timestamp
        self.model_type = model_type
        self.regression_model = regression_model
        self.logger = logger
        self.data = self.data.sort_index()
        self.data = self.data[~self.data.index.duplicated(keep="first")]
        self.data_exo = None
        self.steps = None
        self.model = None
        self.grid_search = None

    @staticmethod
    def add_date_features(data: pd.DataFrame, date_features: list, timestamp: str) -> pd.DataFrame:
        """Add date features from the input DataFrame timestamp.

        :param data: The input DataFrame
        :type data: pd.DataFrame
        :param timestamp: The column containing the timestamp
        :type timestamp: str
        :return: The DataFrame with the added features
        :rtype: pd.DataFrame
        """
        df = copy.deepcopy(data)  # noqa: PD901
        df[timestamp] = pd.to_datetime(df["timestamp"])
        if "year" in date_features:
            df["year"] = [i.year for i in df["timestamp"]]
        if "month" in date_features:
            df["month"] = [i.month for i in df["timestamp"]]
        if "day_of_week" in date_features:
            df["day_of_week"] = [i.dayofweek for i in df["timestamp"]]
        if "day_of_year" in date_features:
            df["day_of_year"] = [i.dayofyear for i in df["timestamp"]]
        if "day" in date_features:
            df["day"] = [i.day for i in df["timestamp"]]
        if "hour" in date_features:
            df["hour"] = [i.day for i in df["timestamp"]]
        return df

    def get_regression_model(self: MLRegressor) -> tuple[str, str]:
        r"""
        Get the base model and parameter grid for the specified regression model.
        Returns a tuple containing the base model and parameter grid corresponding to \
            the specified regression model.

        :param self: The instance of the MLRegressor class.
        :type self: MLRegressor
        :return: A tuple containing the base model and parameter grid.
        :rtype: tuple[str, str]
        """
        if self.regression_model == "LinearRegression":
            base_model = REGRESSION_METHODS["LinearRegression"]["model"]
            param_grid = REGRESSION_METHODS["LinearRegression"]["param_grid"]
        elif self.regression_model == "RidgeRegression":
            base_model = REGRESSION_METHODS["RidgeRegression"]["model"]
            param_grid = REGRESSION_METHODS["RidgeRegression"]["param_grid"]
        elif self.regression_model == "LassoRegression":
            base_model = REGRESSION_METHODS["LassoRegression"]["model"]
            param_grid = REGRESSION_METHODS["LassoRegression"]["param_grid"]
        elif self.regression_model == "RandomForestRegression":
            base_model = REGRESSION_METHODS["RandomForestRegression"]["model"]
            param_grid = REGRESSION_METHODS["RandomForestRegression"]["param_grid"]
        elif self.regression_model == "GradientBoostingRegression":
            base_model = REGRESSION_METHODS["GradientBoostingRegression"]["model"]
            param_grid = REGRESSION_METHODS["GradientBoostingRegression"]["param_grid"]
        elif self.regression_model == "AdaBoostRegression":
            base_model = REGRESSION_METHODS["AdaBoostRegression"]["model"]
            param_grid = REGRESSION_METHODS["AdaBoostRegression"]["param_grid"]
        else:
            self.logger.error(
                "Passed model %s is not valid",
                self.regression_model,
            )
            return None
        return base_model, param_grid

    def fit(self: MLRegressor, date_features: list | None = None) -> None:
        r"""Fit the model using the provided data.

        :param date_features: A list of 'date_features' to take into account when \
            fitting the model.
        :type data: list
        """
        self.logger.info("Performing a MLRegressor fit for %s", self.model_type)
        self.data_exo = pd.DataFrame(self.data)
        self.data_exo[self.features] = self.data[self.features]
        self.data_exo[self.target] = self.data[self.target]
        keep_columns = []
        keep_columns.extend(self.features)
        if self.timestamp is not None:
            keep_columns.append(self.timestamp)
        keep_columns.append(self.target)
        self.data_exo = self.data_exo[self.data_exo.columns.intersection(keep_columns)]
        self.data_exo = self.data_exo.reset_index(drop=True)
        if date_features is not None:
            if self.timestamp is not None:
                self.data_exo = MLRegressor.add_date_features(
                    self.data_exo,
                    date_features,
                    self.timestamp,
                )
            else:
                self.logger.error(
                    "If no timestamp provided, you can't use date_features, going \
                    further without date_features.",
                )
        y = self.data_exo[self.target]
        self.data_exo = self.data_exo.drop(self.target, axis=1)
        if self.timestamp is not None:
            self.data_exo = self.data_exo.drop(self.timestamp, axis=1)
        X = self.data_exo 
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
        self.steps = len(X_test)
        base_model, param_grid = self.get_regression_model()
        self.model = make_pipeline(StandardScaler(), base_model)
        # Create a grid search object
        self.grid_search = GridSearchCV(self.model, param_grid, cv=5, scoring="neg_mean_squared_error",
                                        refit=True, verbose=0, n_jobs=-1)
        # Fit the grid search object to the data
        self.logger.info("Training a %s model", self.regression_model)
        start_time = time.time()
        self.grid_search.fit(X_train.values, y_train.values)
        self.logger.info("Elapsed time for model fit: %s", time.time() - start_time)
        self.model = self.grid_search.best_estimator_
        # Make predictions
        predictions = self.model.predict(X_test.values)
        predictions = pd.Series(predictions, index=X_test.index)
        pred_metric = r2_score(y_test, predictions)
        self.logger.info(
            "Prediction R2 score of fitted model on test data: %s",
            pred_metric,
        )

    def predict(self: MLRegressor, new_values: list) -> np.ndarray:
        """Predict a new value.

        :param new_values: The new values for the features \
            (in the same order as the features list). \
            Example: [2.24, 5.68].
        :type new_values: list
        :return: The np.ndarray containing the predicted value.
        :rtype: np.ndarray
        """
        self.logger.info("Performing a prediction for %s", self.model_type)
        new_values = np.array([new_values])
        return self.model.predict(new_values)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import copy
import time
from typing import Optional, Tuple
import pandas as pd
import numpy as np

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import ElasticNet
from sklearn.neighbors import KNeighborsRegressor
from sklearn.metrics import r2_score

from skforecast.ForecasterAutoreg import ForecasterAutoreg
from skforecast.model_selection import bayesian_search_forecaster
from skforecast.model_selection import backtesting_forecaster

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

class MLForecaster:
    r"""
    A forecaster class using machine learning models with auto-regressive approach and features\
    based on timestamp information (hour, day, week, etc).
    
    This class uses the `skforecast` module and the machine learning models are from `scikit-learn`.
    
    It exposes three main methods:
    
    - `fit`: to train a model with the passed data.
    
    - `predict`: to obtain a forecast from a pre-trained model.
    
    - `tune`: to optimize the models hyperparameters using bayesian optimization. 
    
    """

    def __init__(self, data: pd.DataFrame, model_type: str, var_model: str, sklearn_model: str,
                 num_lags: int, emhass_conf: dict, logger: logging.Logger) -> None:
        r"""Define constructor for the forecast class.

        :param data: The data that will be used for train/test
        :type data: pd.DataFrame
        :param model_type: A unique name defining this model and useful to identify \
            for what it will be used for.
        :type model_type: str
        :param var_model: The name of the sensor to retrieve data from Home Assistant. \
            Example: `sensor.power_load_no_var_loads`.
        :type var_model: str
        :param sklearn_model: The `scikit-learn` model that will be used. For now only \
            this options are possible: `LinearRegression`, `ElasticNet` and `KNeighborsRegressor`.
        :type sklearn_model: str
        :param num_lags: The number of auto-regression lags to consider. A good starting point \
            is to fix this as one day. For example if your time step is 30 minutes, then fix this \
            to 48, if the time step is 1 hour the fix this to 24 and so on.
        :type num_lags: int
        :param emhass_conf: Dictionary containing the needed emhass paths
        :type emhass_conf: dict
        :param logger: The passed logger object
        :type logger: logging.Logger
        """
        self.data = data
        self.model_type = model_type
        self.var_model = var_model
        self.sklearn_model = sklearn_model
        self.num_lags = num_lags
        self.emhass_conf = emhass_conf
        self.logger = logger
        self.is_tuned = False
        # A quick data preparation
        self.data.index = pd.to_datetime(self.data.index)
        self.data.sort_index(inplace=True)
        self.data = self.data[~self.data.index.duplicated(keep='first')]
    
    @staticmethod
    def add_date_features(data: pd.DataFrame) -> pd.DataFrame:
        """Add date features from the input DataFrame timestamp

        :param data: The input DataFrame
        :type data: pd.DataFrame
        :return: The DataFrame with the added features
        :rtype: pd.DataFrame
        """
        df = copy.deepcopy(data)
        df['year'] = [i.year for i in df.index]
        df['month'] = [i.month for i in df.index]
        df['day_of_week'] = [i.dayofweek for i in df.index]
        df['day_of_year'] = [i.dayofyear for i in df.index]
        df['day'] = [i.day for i in df.index]
        df['hour'] = [i.hour for i in df.index]
        return df

    @staticmethod
    def neg_r2_score(y_true, y_pred):
        """The negative of the r2 score."""
        return -r2_score(y_true, y_pred)
    
    @staticmethod
    def generate_exog(data_last_window, periods, var_name):
        """Generate the exogenous data for future timestamps."""
        forecast_dates = pd.date_range(start=data_last_window.index[-1]+data_last_window.index.freq, 
                                       periods=periods, 
                                       freq=data_last_window.index.freq)
        exog = pd.DataFrame({var_name:[np.nan]*periods},
                            index=forecast_dates)
        exog = MLForecaster.add_date_features(exog)
        return exog
    
    def fit(self, split_date_delta: Optional[str] = '48h', perform_backtest: Optional[bool] = False
            ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        r"""The fit method to train the ML model.

        :param split_date_delta: The delta from now to `split_date_delta` that will be used \
            as the test period to evaluate the model, defaults to '48h'
        :type split_date_delta: Optional[str], optional
        :param perform_backtest: If `True` then a back testing routine is performed to evaluate \
            the performance of the model on the complete train set, defaults to False
        :type perform_backtest: Optional[bool], optional
        :return: The DataFrame containing the forecast data results without and with backtest
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]
        """
        self.logger.info("Performing a forecast model fit for "+self.model_type)
        # Preparing the data: adding exogenous features
        self.data_exo = pd.DataFrame(index=self.data.index)
        self.data_exo = MLForecaster.add_date_features(self.data_exo)
        self.data_exo[self.var_model] = self.data[self.var_model]
        self.data_exo = self.data_exo.interpolate(method='linear', axis=0, limit=None)
        # train/test split
        self.date_train = self.data_exo.index[-1]-pd.Timedelta('5days')+self.data_exo.index.freq # The last 5 days
        self.date_split = self.data_exo.index[-1]-pd.Timedelta(split_date_delta)+self.data_exo.index.freq # The last 48h
        self.data_train = self.data_exo.loc[:self.date_split-self.data_exo.index.freq,:]
        self.data_test  = self.data_exo.loc[self.date_split:,:]
        self.steps = len(self.data_test)
        # Pick correct sklearn model
        if self.sklearn_model == 'LinearRegression':
            base_model = LinearRegression()
        elif self.sklearn_model == 'ElasticNet':
            base_model = ElasticNet()
        elif self.sklearn_model == 'KNeighborsRegressor':
            base_model = KNeighborsRegressor()
        else:
            self.logger.error("Passed sklearn model "+self.sklearn_model+" is not valid")
        # Define the forecaster object
        self.forecaster = ForecasterAutoreg(
            regressor = base_model,
            lags      = self.num_lags
            )
        # Fit and time it
        self.logger.info("Training a "+self.sklearn_model+" model")
        start_time = time.time()
        self.forecaster.fit(y=self.data_train[self.var_model], 
                            exog=self.data_train.drop(self.var_model, axis=1))
        self.logger.info(f"Elapsed time for model fit: {time.time() - start_time}")
        # Make a prediction to print metrics
        predictions = self.forecaster.predict(steps=self.steps, exog=self.data_test.drop(self.var_model, axis=1))
        pred_metric = r2_score(self.data_test[self.var_model],predictions)
        self.logger.info(f"Prediction R2 score of fitted model on test data: {pred_metric}")
        # Packing results in a DataFrame
        df_pred = pd.DataFrame(index=self.data_exo.index,columns=['train','test','pred'])
        df_pred['train'] = self.data_train[self.var_model]
        df_pred['test'] = self.data_test[self.var_model]
        df_pred['pred'] = predictions
        df_pred_backtest = None
        if perform_backtest is True:
            # Using backtesting tool to evaluate the model
            self.logger.info("Performing simple backtesting of fitted model")
            start_time = time.time()
            metric, predictions_backtest = backtesting_forecaster(
                forecaster         = self.forecaster,
                y                  = self.data_train[self.var_model],
                exog               = self.data_train.drop(self.var_model, axis=1),
                initial_train_size = None,
                fixed_train_size   = False,
                steps              = self.num_lags,
                metric             = MLForecaster.neg_r2_score,
                refit              = False,
                verbose            = False
            )
            self.logger.info(f"Elapsed backtesting time: {time.time() - start_time}")
            self.logger.info(f"Backtest R2 score: {-metric}")
            df_pred_backtest = pd.DataFrame(index=self.data_exo.index,columns=['train','pred'])
            df_pred_backtest['train'] = self.data_exo[self.var_model]
            df_pred_backtest['pred'] = predictions_backtest
        return df_pred, df_pred_backtest
    
    def predict(self, data_last_window: Optional[pd.DataFrame] = None
            ) -> pd.Series:
        """The predict method to generate forecasts from a previously fitted ML model.

        :param data_last_window: The data that will be used to generate the new forecast, this \
            will be freshly retrieved from Home Assistant. This data is needed because the forecast \
            model is an auto-regressive model with lags. If not passed then the data used during the \
            model train is used, defaults to None
        :type data_last_window: Optional[pd.DataFrame], optional
        :return: A pandas series containing the generated forecasts.
        :rtype: pd.Series
        """
        if data_last_window is None:
            predictions = self.forecaster.predict(steps=self.num_lags, exog=self.data_test.drop(self.var_model, axis=1))
        else:
            data_last_window = data_last_window.interpolate(method='linear', axis=0, limit=None)
            if self.is_tuned:
                exog = MLForecaster.generate_exog(data_last_window, self.lags_opt, self.var_model)
                predictions = self.forecaster.predict(steps=self.lags_opt, 
                                                      last_window=data_last_window[self.var_model],
                                                      exog=exog.drop(self.var_model, axis=1))
            else:
                exog = MLForecaster.generate_exog(data_last_window, self.num_lags, self.var_model)
                predictions = self.forecaster.predict(steps=self.num_lags, 
                                                      last_window=data_last_window[self.var_model],
                                                      exog=exog.drop(self.var_model, axis=1))
        return predictions
    
    def tune(self, debug: Optional[bool] = False) -> pd.DataFrame:
        """Tuning a previously fitted model using bayesian optimization.

        :param debug: Set to True for testing and faster optimizations, defaults to False
        :type debug: Optional[bool], optional
        :return: The DataFrame with the forecasts using the optimized model.
        :rtype: pd.DataFrame
        """
        # Regressor hyperparameters search space
        if self.sklearn_model == 'LinearRegression':
            if debug:
                def search_space(trial):
                    search_space  = {'fit_intercept': trial.suggest_categorical('fit_intercept', [True]),
                                     'lags': trial.suggest_categorical('lags', [3])} 
                    return search_space
            else:
                def search_space(trial):
                    search_space  = {'fit_intercept': trial.suggest_categorical('fit_intercept', [True, False]),
                                     'lags': trial.suggest_categorical('lags', [6, 12, 24, 36, 48, 60, 72])}
                    return search_space
        elif self.sklearn_model == 'ElasticNet':
            if debug:
                def search_space(trial):
                    search_space  = {'selection': trial.suggest_categorical('selection', ['random']),
                                     'lags': trial.suggest_categorical('lags', [3])} 
                    return search_space
            else:
                def search_space(trial):
                    search_space  = {'alpha': trial.suggest_float('alpha', 0.0, 2.0),
                                    'l1_ratio': trial.suggest_float('l1_ratio', 0.0, 1.0),
                                    'selection': trial.suggest_categorical('selection', ['cyclic', 'random']),
                                    'lags': trial.suggest_categorical('lags', [6, 12, 24, 36, 48, 60, 72])} 
                    return search_space
        elif self.sklearn_model == 'KNeighborsRegressor':
            if debug:
                def search_space(trial):
                    search_space  = {'weights': trial.suggest_categorical('weights', ['uniform']),
                                     'lags': trial.suggest_categorical('lags', [3])} 
                    return search_space
            else:
                def search_space(trial):
                    search_space  = {'n_neighbors': trial.suggest_int('n_neighbors', 2, 20),
                                    'leaf_size': trial.suggest_int('leaf_size', 20, 40),
                                    'weights': trial.suggest_categorical('weights', ['uniform', 'distance']),
                                    'lags': trial.suggest_categorical('lags', [6, 12, 24, 36, 48, 60, 72])} 
                    return search_space
        # Bayesian search hyperparameter and lags with skforecast/optuna
        # Lags used as predictors
        if debug:
            refit = False
            num_lags = 3
        else:
            refit = True
            num_lags = self.num_lags
        # The optimization routine call
        self.logger.info("Bayesian hyperparameter optimization with backtesting")
        start_time = time.time()
        self.optimize_results, self.optimize_results_object = bayesian_search_forecaster(
            forecaster         = self.forecaster,
            y                  = self.data_train[self.var_model],
            exog               = self.data_train.drop(self.var_model, axis=1),
            search_space       = search_space,
            steps              = num_lags,
            metric             = MLForecaster.neg_r2_score,
            refit              = refit,
            initial_train_size = len(self.data_exo.loc[:self.date_train]),
            fixed_train_size   = True,
            n_trials           = 10,
            random_state       = 123,
            return_best        = True,
            verbose            = False,
            engine             = 'optuna'
        )
        self.logger.info(f"Elapsed time: {time.time() - start_time}")
        self.is_tuned = True
        predictions_opt = self.forecaster.predict(steps=self.num_lags, exog=self.data_test.drop(self.var_model, axis=1))
        freq_hours = self.data_exo.index.freq.delta.seconds/3600
        self.lags_opt = int(np.round(len(self.optimize_results.iloc[0]['lags'])))
        self.days_needed = int(np.round(self.lags_opt*freq_hours/24))
        df_pred_opt = pd.DataFrame(index=self.data_exo.index,columns=['train','test','pred_optim'])
        df_pred_opt['train'] = self.data_train[self.var_model]
        df_pred_opt['test'] = self.data_test[self.var_model]
        df_pred_opt['pred_optim'] = predictions_opt
        pred_optim_metric_train = -self.optimize_results.iloc[0]['neg_r2_score']
        self.logger.info(f"R2 score for optimized prediction in train period: {pred_optim_metric_train}")
        pred_optim_metric_test = r2_score(df_pred_opt.loc[predictions_opt.index,'test'],
                                          df_pred_opt.loc[predictions_opt.index,'pred_optim'])
        self.logger.info(f"R2 score for optimized prediction in test period: {pred_optim_metric_test}")
        self.logger.info("Number of optimal lags obtained: "+str(self.lags_opt))
        return df_pred_opt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import pandas as pd
import pathlib, json, yaml, copy

from emhass import utils

# the root folder
root = str(utils.get_root(__file__, num_parent=2))
emhass_conf = {}
emhass_conf['config_path'] = pathlib.Path(root) / 'config_emhass.yaml'
emhass_conf['data_path'] = pathlib.Path(root) / 'data/'
emhass_conf['root_path'] = pathlib.Path(root) / 'src/emhass/'

# create logger
logger, ch = utils.get_logger(__name__, emhass_conf, save_to_file=False)

class TestCommandLineUtils(unittest.TestCase):
    
    @staticmethod
    def get_test_params():
        with open(emhass_conf['config_path'], 'r') as file:
            params = yaml.load(file, Loader=yaml.FullLoader)
        params.update({
            'params_secrets': {
                'hass_url': 'http://supervisor/core/api',
                'long_lived_token': '${SUPERVISOR_TOKEN}',
                'time_zone': 'Europe/Paris',
                'lat': 45.83,
                'lon': 6.86,
                'alt': 8000.0
            }
            })
        return params

    def setUp(self):
        params = TestCommandLineUtils.get_test_params()
        runtimeparams = {
            'pv_power_forecast':[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48],
            'load_power_forecast':[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48],
            'load_cost_forecast':[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48],
            'prod_price_forecast':[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48]
        }
        self.runtimeparams_json = json.dumps(runtimeparams)
        params['passed_data'] = runtimeparams
        params['optim_conf']['weather_forecast_method'] = 'list'
        params['optim_conf']['load_forecast_method'] = 'list'
        params['optim_conf']['load_cost_forecast_method'] = 'list'
        params['optim_conf']['prod_price_forecast_method'] = 'list'
        self.params_json = json.dumps(params)
        
    def test_get_yaml_parse(self):
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=False)
        self.assertIsInstance(retrieve_hass_conf, dict)
        self.assertIsInstance(optim_conf, dict)
        self.assertIsInstance(plant_conf, dict)
        self.assertTrue(retrieve_hass_conf['alt'] == 4807.8)
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=self.params_json)
        self.assertTrue(retrieve_hass_conf['alt'] == 8000.0)
        
    def test_get_forecast_dates(self):
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=self.params_json)
        freq = int(retrieve_hass_conf['freq'].seconds/60.0)
        delta_forecast = int(optim_conf['delta_forecast'].days)
        forecast_dates = utils.get_forecast_dates(freq, delta_forecast)
        self.assertIsInstance(forecast_dates, pd.core.indexes.datetimes.DatetimeIndex)
        self.assertTrue(len(forecast_dates)==int(delta_forecast*60*24/freq))
        
    def test_treat_runtimeparams(self):
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=self.params_json)
        set_type = 'dayahead-optim'
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            self.runtimeparams_json, self.params_json,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        self.assertIsInstance(params, str)
        params = json.loads(params)
        self.assertIsInstance(params['passed_data']['pv_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_cost_forecast'], list)
        self.assertIsInstance(params['passed_data']['prod_price_forecast'], list)
        self.assertTrue(optim_conf['weather_forecast_method'] == 'list')
        self.assertTrue(optim_conf['load_forecast_method'] == 'list')
        self.assertTrue(optim_conf['load_cost_forecast_method'] == 'list')
        self.assertTrue(optim_conf['prod_price_forecast_method'] == 'list')
        set_type = 'naive-mpc-optim'
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            self.runtimeparams_json, self.params_json,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        self.assertIsInstance(params, str)
        params = json.loads(params)
        self.assertTrue(params['passed_data']['prediction_horizon'] == 10)
        self.assertTrue(params['passed_data']['soc_init'] == plant_conf['SOCtarget'])
        self.assertTrue(params['passed_data']['soc_final'] == plant_conf['SOCtarget'])
        self.assertTrue(params['passed_data']['def_total_hours'] == optim_conf['def_total_hours'])
        # This will be the case when using emhass in standalone mode
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=self.params_json)
        params = json.dumps(None)
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            self.runtimeparams_json, params,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        self.assertIsInstance(params, str)
        params = json.loads(params)
        self.assertIsInstance(params['passed_data']['pv_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_cost_forecast'], list)
        self.assertIsInstance(params['passed_data']['prod_price_forecast'], list)
        self.assertTrue(optim_conf['weather_forecast_method'] == 'list')
        self.assertTrue(optim_conf['load_forecast_method'] == 'list')
        self.assertTrue(optim_conf['load_cost_forecast_method'] == 'list')
        self.assertTrue(optim_conf['prod_price_forecast_method'] == 'list')
        # Test passing optimization and plant configuration parameters at runtime 
        runtimeparams = json.loads(self.runtimeparams_json)
        runtimeparams.update({'num_def_loads':3})
        runtimeparams.update({'P_deferrable_nom':[3000.0, 750.0, 2500.0]})
        runtimeparams.update({'def_total_hours':[5, 8, 10]})
        runtimeparams.update({'treat_def_as_semi_cont':[True, True, True]})
        runtimeparams.update({'set_def_constant':[False, False, False]})
        runtimeparams.update({'weight_battery_discharge':2.0})
        runtimeparams.update({'weight_battery_charge':2.0})
        runtimeparams.update({'solcast_api_key':'yoursecretsolcastapikey'})
        runtimeparams.update({'solcast_rooftop_id':'yourrooftopid'})
        runtimeparams.update({'solar_forecast_kwp':5.0})
        runtimeparams.update({'SOCtarget':0.4})
        runtimeparams.update({'publish_prefix':'emhass_'})
        runtimeparams.update({'custom_pv_forecast_id':'my_custom_pv_forecast_id'})
        runtimeparams.update({'custom_load_forecast_id':'my_custom_load_forecast_id'})
        runtimeparams.update({'custom_batt_forecast_id':'my_custom_batt_forecast_id'})
        runtimeparams.update({'custom_batt_soc_forecast_id':'my_custom_batt_soc_forecast_id'})
        runtimeparams.update({'custom_grid_forecast_id':'my_custom_grid_forecast_id'})
        runtimeparams.update({'custom_cost_fun_id':'my_custom_cost_fun_id'})
        runtimeparams.update({'custom_optim_status_id':'my_custom_optim_status_id'})
        runtimeparams.update({'custom_unit_load_cost_id':'my_custom_unit_load_cost_id'})
        runtimeparams.update({'custom_unit_prod_price_id':'my_custom_unit_prod_price_id'})
        runtimeparams.update({'custom_deferrable_forecast_id':'my_custom_deferrable_forecast_id'})
        runtimeparams_json = json.dumps(runtimeparams)
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=self.params_json)
        set_type = 'dayahead-optim'
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            runtimeparams_json, self.params_json,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        self.assertIsInstance(params, str)
        params = json.loads(params)
        self.assertIsInstance(params['passed_data']['pv_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_power_forecast'], list)
        self.assertIsInstance(params['passed_data']['load_cost_forecast'], list)
        self.assertIsInstance(params['passed_data']['prod_price_forecast'], list)
        self.assertTrue(optim_conf['num_def_loads'] == 3)
        self.assertTrue(optim_conf['P_deferrable_nom'] == [3000.0, 750.0, 2500.0])
        self.assertTrue(optim_conf['def_total_hours'] == [5, 8, 10])
        self.assertTrue(optim_conf['treat_def_as_semi_cont'] == [True, True, True])
        self.assertTrue(optim_conf['set_def_constant'] == [False, False, False])
        self.assertTrue(optim_conf['weight_battery_discharge'] == 2.0)
        self.assertTrue(optim_conf['weight_battery_charge'] == 2.0)
        self.assertTrue(retrieve_hass_conf['solcast_api_key'] == 'yoursecretsolcastapikey')
        self.assertTrue(retrieve_hass_conf['solcast_rooftop_id'] == 'yourrooftopid')
        self.assertTrue(retrieve_hass_conf['solar_forecast_kwp'] == 5.0)
        self.assertTrue(plant_conf['SOCtarget'] == 0.4)
        self.assertTrue(params['passed_data']['publish_prefix'] == 'emhass_')
        self.assertTrue(params['passed_data']['custom_pv_forecast_id'] == 'my_custom_pv_forecast_id')
        self.assertTrue(params['passed_data']['custom_load_forecast_id'] == 'my_custom_load_forecast_id')
        self.assertTrue(params['passed_data']['custom_batt_forecast_id'] == 'my_custom_batt_forecast_id')
        self.assertTrue(params['passed_data']['custom_batt_soc_forecast_id'] == 'my_custom_batt_soc_forecast_id')
        self.assertTrue(params['passed_data']['custom_grid_forecast_id'] == 'my_custom_grid_forecast_id')
        self.assertTrue(params['passed_data']['custom_cost_fun_id'] == 'my_custom_cost_fun_id')
        self.assertTrue(params['passed_data']['custom_optim_status_id'] == 'my_custom_optim_status_id')
        self.assertTrue(params['passed_data']['custom_unit_load_cost_id'] == 'my_custom_unit_load_cost_id')
        self.assertTrue(params['passed_data']['custom_unit_prod_price_id'] == 'my_custom_unit_prod_price_id')
        self.assertTrue(params['passed_data']['custom_deferrable_forecast_id'] == 'my_custom_deferrable_forecast_id')
    
    def test_treat_runtimeparams_failed(self):
        params = TestCommandLineUtils.get_test_params()
        runtimeparams = {
            'pv_power_forecast':[1,2,3,4,5,'nan',7,8,9,10],
            'load_power_forecast':[1,2,'nan',4,5,6,7,8,9,10],
            'load_cost_forecast':[1,2,3,4,5,6,7,8,'nan',10],
            'prod_price_forecast':[1,2,3,4,'nan',6,7,8,9,10]
        }
        runtimeparams_json = json.dumps(runtimeparams)
        params['passed_data'] = runtimeparams
        params['optim_conf']['weather_forecast_method'] = 'list'
        params['optim_conf']['load_forecast_method'] = 'list'
        params['optim_conf']['load_cost_forecast_method'] = 'list'
        params['optim_conf']['prod_price_forecast_method'] = 'list'
        params_json = json.dumps(params)
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=params_json)
        set_type = 'dayahead-optim'
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            runtimeparams_json, params_json,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        params = json.loads(params)
        runtimeparams = json.loads(runtimeparams_json)
        self.assertTrue(len([x for x in runtimeparams['pv_power_forecast'] if not str(x).isdigit()])>0)
        self.assertTrue(len([x for x in runtimeparams['load_power_forecast'] if not str(x).isdigit()])>0)
        self.assertTrue(len([x for x in runtimeparams['load_cost_forecast'] if not str(x).isdigit()])>0)
        self.assertTrue(len([x for x in runtimeparams['prod_price_forecast'] if not str(x).isdigit()])>0)
        params = TestCommandLineUtils.get_test_params()
        runtimeparams = {
            'pv_power_forecast':'[1,2,3,4,5,6,7,8,9,10]',
            'load_power_forecast':'[1,2,3,4,5,6,7,8,9,10]',
            'load_cost_forecast':'[1,2,3,4,5,6,7,8,9,10]',
            'prod_price_forecast':'[1,2,3,4,5,6,7,8,9,10]'
        }
        runtimeparams_json = json.dumps(runtimeparams)
        params['passed_data'] = runtimeparams
        params['optim_conf']['weather_forecast_method'] = 'list'
        params['optim_conf']['load_forecast_method'] = 'list'
        params['optim_conf']['load_cost_forecast_method'] = 'list'
        params['optim_conf']['prod_price_forecast_method'] = 'list'
        params_json = json.dumps(params)
        retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(emhass_conf, use_secrets=True, params=params_json)
        set_type = 'dayahead-optim'
        params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
            runtimeparams_json, params_json,
            retrieve_hass_conf, optim_conf, plant_conf, set_type, logger)
        params = json.loads(params)
        runtimeparams = json.loads(runtimeparams_json)
        self.assertIsInstance(runtimeparams['pv_power_forecast'], str)
        self.assertIsInstance(runtimeparams['load_power_forecast'], str)
        self.assertIsInstance(runtimeparams['load_cost_forecast'], str)
        self.assertIsInstance(runtimeparams['prod_price_forecast'], str)
    
    def test_build_params(self):
        with open(emhass_conf['config_path'], 'r') as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
        retrieve_hass_conf = config['retrieve_hass_conf']
        optim_conf = config['optim_conf']
        plant_conf = config['plant_conf']
        params = {}
        params['retrieve_hass_conf'] = retrieve_hass_conf
        params['optim_conf'] = optim_conf
        params['plant_conf'] = plant_conf
        options_json = emhass_conf['config_path'].parent / "options.json"
        # Read options info
        with options_json.open('r') as data:
            options = json.load(data)
        with open(emhass_conf['config_path'].parent / "secrets_emhass(example).yaml", 'r') as file:
            params_secrets = yaml.load(file, Loader=yaml.FullLoader)
        addon = 1
        params = utils.build_params(params, params_secrets, options, addon, logger)
        expected_keys = ['retrieve_hass_conf', 'params_secrets', 'optim_conf', 'plant_conf', 'passed_data']
        for key in expected_keys:
            self.assertTrue(key in params.keys())
        self.assertTrue(params['params_secrets']['time_zone'] == "Europe/Paris")
        params = {}
        params['retrieve_hass_conf'] = retrieve_hass_conf
        params['optim_conf'] = optim_conf
        params['plant_conf'] = plant_conf
        addon = 0
        params = utils.build_params(params, params_secrets, options, addon, logger)
        for key in expected_keys:
            self.assertTrue(key in params.keys())
        self.assertTrue(params['params_secrets']['time_zone'] == "Europe/Paris")
        
if __name__ == '__main__':
    unittest.main()
    ch.close()
    logger.removeHandler(ch)